from particleShearLinkableObjects import EnsembleCompactParticlesAdjustableInterfaceStrengthFromModelParameters
from particleShearObjects import particle_shear_model_parameters
from .OscillatoryShearExperiment import OscillatoryShearExperiment
import math
from tkinter import *

from .Simulation_dermal_filler_rheology import Simulation_dermal_filler_rheology



class Simulation_interlocking_rheology(Simulation_dermal_filler_rheology):

    def __init__(self,root_folder,do_permanent_links=True,cut_lines=5,N=150,packing_fraction=2,
                 bimodal_factor=1.4,
                 amplitude=0.5,
                 Young_modulus_spheres=1000,
                 density=50,
                 mu=0.1,relative_viscosity=0.01,
                 relative_y_scale_force=5e-4,relative_frequency=0.05, relative_transversal_link_strength=1,
                 avoid_horizontal_angle_degree=15,avoid_height_spanning_particles=False,
                 interface_reenforcement_central=1,interface_reenforcement_tangential=1,
                 central_repulsion_coefficient=1,
                 keep_viscosity_coefficients_constant=True,cut_top_bottom=True,
                 doCutByTriangulation=True,remove_link_fraction=0,edge_fuzziness=0,doDrawing=True,saveData=False):



        self.avoid_horizontal_angle_degree=avoid_horizontal_angle_degree
        self.interface_reenforcement_central=interface_reenforcement_central
        self.interface_reenforcement_tangential=interface_reenforcement_tangential
        self.keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant
        self.cut_top_bottom=cut_top_bottom
        self.remove_link_fraction=remove_link_fraction
        self.edge_fuzziness=edge_fuzziness


        super(Simulation_interlocking_rheology,self).__init__(
            root_folder,do_permanent_links=do_permanent_links,cut_lines=cut_lines,
            N=N,packing_fraction=packing_fraction,
            bimodal_factor=bimodal_factor,
            amplitude=amplitude,
            Young_modulus_spheres=Young_modulus_spheres,
            density=density,
            mu=mu,relative_viscosity=relative_viscosity,
            relative_y_scale_force=relative_y_scale_force,relative_frequency=relative_frequency,
            relative_transversal_link_strength=relative_transversal_link_strength,
            central_repulsion_coefficient=central_repulsion_coefficient,
            avoid_horizontal_angle_degree=avoid_horizontal_angle_degree,
            avoid_height_spanning_particles=avoid_height_spanning_particles,
            doCutByTriangulation=doCutByTriangulation,
            doDrawing=doDrawing,saveData=saveData)



        self.function_call = "Simulation_interlocking_rheology(root_folder=" + str(root_folder) + ",do_permanent_links=" + \
                             str(do_permanent_links) + ",cut_lines=" + str(cut_lines) + ",N=" + str(N) + \
                             ",packing_fraction=" + str(packing_fraction) + \
                             ",bimodal_factor=" + str(bimodal_factor) + \
                             ",amplitude=" + str(amplitude) + \
                             ",Young_modulus_spheres=" + str(Young_modulus_spheres)  + \
                             ",density=" + str(density) + ",mu=" + str(mu) +  \
                             ",relative_viscosity="+str(relative_viscosity)+\
                             ",relative_y_scale_force=" + str(relative_y_scale_force) + ",relative_frequency=" + \
                             str(relative_frequency) +\
                             ",relative_transversal_link_strength="+str(relative_transversal_link_strength)+ \
                             ",central_repulsion_coefficient=" + str(central_repulsion_coefficient) + \
                             ",avoid_horizontal_angle_degree="+str(avoid_horizontal_angle_degree)+ \
                             ",avoid_height_spanning_particles=" + str(self.avoid_height_spanning_particles) + \
                             ",interface_reenforcement_central="+str(interface_reenforcement_central)+\
                             ",interface_reenforcement_tangential="+str(interface_reenforcement_tangential)+\
                             ",keep_viscosity_coefficients_constant="+str(keep_viscosity_coefficients_constant)+ \
                             ",cut_top_bottom=" + str(cut_top_bottom)+ \
                             ",doCutByTriangulation=" + str(self.doCutByTriangulation) + \
                             ",remove_link_fraction=" + str(self.remove_link_fraction) + \
                             ",edge_fuzziness=" + str(self.edge_fuzziness) + \
                             ",doDrawing=" + str(doDrawing)+")"







    def initEnsemble(self,do_debug=False):
        super(Simulation_dermal_filler_rheology,self).initEnsemble()

        print("Simulation_interlocking_rheology: initiating ensemble")



        self.theEnsemble = EnsembleCompactParticlesAdjustableInterfaceStrengthFromModelParameters(
            cut_lines=self.cut_lines,N=self.N,packing_fraction=self.packing_fraction,
            Young_modulus_spheres=self.Young_modulus_spheres,density=self.density,
            bimodal_factor=self.bimodal_factor,
            do_permanent_links=self.do_permanent_links,mu=self.mu,theTk=self.theTkSimulation,do_pre_equilibration=True,
            relative_viscosity=self.relative_viscosity,central_repulsion_coefficient=self.central_repulsion_coefficient,
            anticipated_amplitude=self.amplitude,relative_transversal_link_strength=self.relative_transversal_link_strength,
            avoid_horizontal_angle_degree=self.avoid_horizontal_angle_degree,
            avoid_height_spanning_particles=self.avoid_height_spanning_particles,
            interface_reenforcement_central=self.interface_reenforcement_central,
            interface_reenforcement_tangential=self.interface_reenforcement_tangential,
            keep_viscosity_coefficients_constant=self.keep_viscosity_coefficients_constant,
            cut_top_bottom=self.cut_top_bottom,doCutByTriangulation=self.doCutByTriangulation,
            remove_link_fraction=self.remove_link_fraction,
            edge_fuzziness=self.edge_fuzziness,
            doDrawing=self.doDrawing,
            do_debug=do_debug
            )

        self.dt=self.theEnsemble.dt
        self.dt_max = self.theEnsemble.dt_max













from tkinter import *
import random
import math
from .SphereLinkableAdjustableInterfaceStrength import SphereLinkableAdjustableInterfaceStrength
from .EnsembleCompactParticles import EnsembleCompactParticles
from .CanvasPointsLinkable import CanvasPointsLinkable
from particleShearObjects import r_estimate
from particleShearObjects import adjust_m
from particleShearObjects import particle_shear_model_parameters


class EnsembleCompactParticlesAdjustableInterfaceStrength(EnsembleCompactParticles):

    def __init__(self, size_x, size_y, N, packing_fraction=0.8, theCanvas=False, doDrawing=False,
                 k=1, nu=0.01,m=1,
                 bimodal_factor=1.4,
                 k_t=1,nu_t=0.01,mu=0.1,dt=1,dt_max=10,theTkSimulation=False,avoid_horizontal_angle_degree=0,
                 permanent_ratio_central=1, permanent_ratio_tangential=1, keep_viscosity_coefficients_constant=True):


        # The idea here is that we can specifically increase the central or tangential forces for the crosslinked particles
        # as compared to the normal frictional contact

        self.permanent_ratio_central=permanent_ratio_central
        self.permanent_ratio_tangential=permanent_ratio_tangential
        self.keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant

        self.cut_top_bottom=False





        super(EnsembleCompactParticlesAdjustableInterfaceStrength, self).__init__(
            size_x, size_y,N=0,packing_fraction=packing_fraction, theCanvas=theCanvas, doDrawing=False,
                 k=k, nu=nu,m=m,bimodal_factor=bimodal_factor,
                 k_t=k_t,nu_t=nu_t,mu=mu,dt=dt,
                 dt_max=dt_max,theTkSimulation=theTkSimulation,avoid_horizontal_angle_degree=avoid_horizontal_angle_degree)



        self.N = N

        self.sphereList = []
        self.doDrawing = doDrawing

        r=self.size_y/2

        if N > 0:
            r = r_estimate(self.size_x, self.size_y, packing_fraction, N,
                           bimodal_upper=self.bimodal_upper,
                           bimodal_lower=self.bimodal_lower)

        self.r_max = 0

        for i in range(N):
            r_sphere = self.random_radius_bimodal(r)
            if r_sphere > self.r_max:
                self.r_max = r_sphere
            self.sphereList.append(SphereLinkableAdjustableInterfaceStrength("grey",
                                                  random.randrange(0, size_x),
                                                  random.randrange(0, size_y), r_sphere * 2,
                                                  adjust_m(m, r_sphere, r),
                                                  i, theCanvas, False, self, self.size_x, self.size_y,
                                                  permanent_ratio_central=self.permanent_ratio_central,
                                                  permanent_ratio_tangential=self.permanent_ratio_tangential,
                                                  keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant
                                                  ))
        self.setShearInSpheres()
        self.setShearRateInSpheres()

        self.set_graphical_output_configuration(
            self.graphical_output_configuration)  # To propagate it also to the spheres

        self.initiateGraphics()




    def particle_info(self):
        return "Frictional spheres under Lees-Edwards boundary conditions, organized in fixed particles, adjustment of" \
               "strength between permament and temporary interfaces\n" \
               "\tSpheres with, bimodal distribution, equal probability, \n\tr1="+\
               str(self.r_max*0.8/1.2)+"r2="+str(self.r_max)+"\n\tidentical mass="+str(self.sphereList[0].m)+"mg"







def EnsembleCompactParticlesAdjustableInterfaceStrengthFromModelParameters(size_x=500,size_y=500,N=100,packing_fraction=1,mu=1,
                                                Young_modulus_spheres=10000,density=50,bimodal_factor=1.4,
                                                cut_lines=5,do_permanent_links=True,theTk=False,do_pre_equilibration=True,
                                                relative_viscosity=0.01,central_repulsion_coefficient=0,anticipated_amplitude=0.1,
                                                relative_transversal_link_strength=1,avoid_horizontal_angle_degree=0,
                                                avoid_height_spanning_particles=False,
                                                interface_reenforcement_central=1,interface_reenforcement_tangential=1,
                                                keep_viscosity_coefficients_constant=True,cut_top_bottom=True,
                                                doCutByTriangulation=True,remove_link_fraction=0,edge_fuzziness=0,doDrawing=True,
                                                do_debug=False):



    

    # First, some logics: with permanent links but insufficient tools to cut them, height spanning particles cannot be avoided

    if avoid_height_spanning_particles and ((doCutByTriangulation and cut_lines <= 1) or ((not doCutByTriangulation)
                                                                                          and cut_Lines <= 0)):
        avoid_height_spanning_particles = False
        print("Cannot avoid height spanning particles for bulk ensembles")

    theCanvas = False

    if doDrawing and not theTk:
        theTk = Tk()

    if doDrawing:
        theCanvas = Canvas(theTk, width=size_x, height=size_y)

        theCanvas.pack()

        theTk.title("Interface: tangential=" + str(interface_reenforcement_tangential) + " central=" + \
                    str(interface_reenforcement_central) + " A=" + str(anticipated_amplitude) + " mu=" + str(mu))




    model = particle_shear_model_parameters(size_x=size_x, size_y=size_y, N=N,
                                            packing_fraction=packing_fraction,
                                            Young_modulus_spheres=Young_modulus_spheres,density=density,
                                            bimodal_factor=1)

    print("Characteristic time constant [s] ",model.time_constant)

    final_dt = 0.1 * model.time_constant/bimodal_factor

    print("dt from model before adapation",final_dt)
    if mu<=1 or relative_transversal_link_strength<1:
        final_dt=final_dt/2
    if  anticipated_amplitude<=0.05:
        final_dt=final_dt/2
    if mu <=0.1 or relative_transversal_link_strength<0.1:
        final_dt = final_dt / 2
    if anticipated_amplitude<=0.005:
        final_dt = final_dt / 2
    if anticipated_amplitude>=0.5:
        final_dt = final_dt / 2
    dt_max = 0.25 * model.time_constant/bimodal_factor



    # Special delicate situations with little viscous components, either central or tangential, and thus
    # tendency for the ensemble to oscillate
    if relative_viscosity<=0.1 and mu<0.1 or relative_transversal_link_strength<0.1 or relative_transversal_link_strength>5:

        dt_max = 0.125 * model.time_constant/bimodal_factor
        print("Relative viscosity below 0.1, adapting dt_max to ",dt_max)
    if relative_viscosity<0.01 and mu<0.01 or relative_transversal_link_strength<0.01:

        dt_max = 0.04 * model.time_constant/bimodal_factor
        print("Relative viscosity below 0.01, adapting dt_max to", dt_max)

    if final_dt > dt_max:
        final_dt=dt_max


    print("dt from model after adaptaion for friction (mu=", mu,") and amplitude (A=",anticipated_amplitude,"):",final_dt)

    if relative_viscosity>1:
        dt_max=dt_max/relative_viscosity

    # Here, the situation is expressed in terms of interface re-enforcement. So since the
    # permanent_ratio_central and permanent_ratio_tangential
    # parameters will be smaller than 1, we need to artificially re-enforce the corresponding k, nu, k_t and nu_t parameters
    # Then, only the values for the permanent links will be multiplied with the smaller constants, and
    # so they reach their intended base level in the permanent links. In the non-permanent (interfacial links),
    # they will remain higher

    nu_for_ensemble=relative_viscosity*model.k * model.time_constant
    # Switch to decide whether viscosity constant should remain globally constant or vary the same way as the elastic ones
    if not keep_viscosity_coefficients_constant:
        nu_for_ensemble=nu_for_ensemble*interface_reenforcement_central

    nu_t_for_ensemble = relative_viscosity*model.k * model.time_constant*relative_transversal_link_strength
    # Switch to decide whether viscosity constant should remain globally constant or vary the same way as the elastic ones
    if not keep_viscosity_coefficients_constant:
        nu_t_for_ensemble = relative_viscosity * model.k * model.time_constant * \
                            relative_transversal_link_strength * interface_reenforcement_tangential



    theEnsemble = EnsembleCompactParticlesAdjustableInterfaceStrength(size_x, size_y, N, packing_fraction,
                                           theCanvas, doDrawing=doDrawing, k=model.k*interface_reenforcement_central,
                                           nu= nu_for_ensemble,
                                           m=model.m,
                                           bimodal_factor=bimodal_factor,
                                           k_t=model.k*relative_transversal_link_strength*interface_reenforcement_tangential,
                                           nu_t=nu_t_for_ensemble ,
                                           mu=mu,
                                           dt=final_dt, dt_max= dt_max, theTkSimulation=theTk,
                                           avoid_horizontal_angle_degree=avoid_horizontal_angle_degree,
                                           permanent_ratio_central=1/interface_reenforcement_central,
                                           permanent_ratio_tangential=1/interface_reenforcement_tangential,
                                           keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant
                                           )



    print("Turning on enhanced central repulsion, coefficient = ",central_repulsion_coefficient)
    theEnsemble.set_central_repulsion_coefficient(central_repulsion_coefficient)

    print("Avoiding full height spanning particles:", avoid_height_spanning_particles)
    theEnsemble.avoid_height_spanning_particles = avoid_height_spanning_particles  # Record this parameter for file saving

    theEnsemble.model=model


    reduction_factor=1

    if(do_debug):
        reduction_factor=10

    if do_pre_equilibration:

        theEnsemble.dt=final_dt

        mdt=0.25*model.time_constant/bimodal_factor
        if relative_viscosity <= 0.1:
            mdt = 0.2 * model.time_constant/bimodal_factor

        theEnsemble.dt_max=mdt

        if theEnsemble.packing_fraction > 2:
            theEnsemble.dt_max = theEnsemble.dt_max / 2


        if theEnsemble.permanent_ratio_central==1:
            theEnsemble.free_pre_equilibration(N=int(300/reduction_factor))
        else:
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))

        if anticipated_amplitude<=0.01 or relative_transversal_link_strength<0.1:
            if theEnsemble.permanent_ratio_central == 1:
                theEnsemble.free_pre_equilibration(N=int(300/reduction_factor))
            else:
                theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))



        theEnsemble.dt = final_dt
        theEnsemble.dt_max = dt_max

        if theEnsemble.permanent_ratio_central == 1:
            theEnsemble.free_pre_equilibration(N=int(200/reduction_factor))
        else:
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))
        if anticipated_amplitude <= 0.01 and theEnsemble.permanent_ratio_central==1:
            theEnsemble.free_pre_equilibration(N=int(200/reduction_factor))
        if anticipated_amplitude <= 0.001 and theEnsemble.permanent_ratio_central==1:
            theEnsemble.free_pre_equilibration(N=int(200/reduction_factor))

    if do_permanent_links:
        theEnsemble.makeAllLinksPermanent()
        if cut_top_bottom:
            theEnsemble.cutTopBottomEdge()
            theEnsemble.cut_top_bottom=True

        if(doCutByTriangulation):
            print("Preparing particles by triangulation; relative fuzziness on edges=",edge_fuzziness)
            theEnsemble.cutByTriangulation(cut_lines,edge_fuzziness=edge_fuzziness)
        else:
            print("Preparing particles by straight line cut")
            for i in range(cut_lines):
                theEnsemble.cutRandomLine()



        max_height=0

        if theEnsemble.avoid_height_spanning_particles:
            max_height=theEnsemble.maximal_particle_height()


        while theEnsemble.avoid_height_spanning_particles and max_height > theEnsemble.size_y*0.9:
            print("Discarded ensemble with full height spanning particles due to max height",max_height)
            print("\twith total particle count", len(theEnsemble.particles))
            theEnsemble.particles=[]  # Reset the particle registry in the particles
            theEnsemble.makeAllLinksPermanent()
            if cut_top_bottom:
                theEnsemble.cutTopBottomEdge()

            if doCutByTriangulation:
                print("Preparing particles by triangulation; relative fuzziness on edges=", edge_fuzziness)
                theEnsemble.cutByTriangulation(cut_lines,edge_fuzziness=edge_fuzziness)
            else:
                print("Preparing particles by straight line cut")
                for i in range(cut_lines):
                    theEnsemble.cutRandomLine()

            max_height = theEnsemble.maximal_particle_height()


        if remove_link_fraction>0:
            theEnsemble.remove_non_essential_links(max_fraction_to_remove=remove_link_fraction,physical_neighbors_only=True)

        theEnsemble.join_isolated_spheres()

        theEnsemble.dt_max = min(theEnsemble.dt*2,theEnsemble.dt_max)
        theEnsemble.dt = theEnsemble.dt

        if relative_transversal_link_strength>5 :
            theEnsemble.dt_max=theEnsemble.dt_max/2



        theEnsemble.free_pre_equilibration(N=int(300/reduction_factor))
        if anticipated_amplitude < 0.1:
            print("Extra equilibration for A<0.1")
            old_mu = theEnsemble.mu
            theEnsemble.mu = 0
            theEnsemble.free_pre_equilibration(N=int(150/reduction_factor))
            theEnsemble.mu = old_mu
            theEnsemble.free_pre_equilibration(N=int(150/reduction_factor))

        if anticipated_amplitude <= 0.01:
            print("Extra equilibration for A<=0.1")
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))
            old_mu = theEnsemble.mu
            theEnsemble.mu = 0
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))
            theEnsemble.mu = old_mu
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))

        if anticipated_amplitude <= 0.001:
            print("Extra equilibration for A<=0.001")
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))
            old_mu = theEnsemble.mu
            theEnsemble.mu = 0
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))
            theEnsemble.mu = old_mu
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))


        if anticipated_amplitude <= 0.0001:
            print("Extra equilibration for A<=0.0001")
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))
            old_mu = theEnsemble.mu
            theEnsemble.mu = 0
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))
            theEnsemble.mu = old_mu
            theEnsemble.free_pre_equilibration(N=int(100/reduction_factor))

        if anticipated_amplitude <= 0.001 and (not (interface_reenforcement_tangential==1) or not (interface_reenforcement_central==1)):
            theEnsemble.dt_max = theEnsemble.dt_max / 3
            print("Extra equilibration for A<=0.001")
            theEnsemble.free_pre_equilibration(N=int(200/reduction_factor))
            old_mu = theEnsemble.mu
            theEnsemble.mu = 0
            theEnsemble.free_pre_equilibration(N=int(200/reduction_factor))
            theEnsemble.mu = old_mu
            theEnsemble.free_pre_equilibration(N=int(200/reduction_factor))

    return theEnsemble






"""Unittest suite for the package particleShear
"""

__all__ = ["TestoneSphereFreeOneSphereBoundary","TestSimulationCrosslinkedShear",
           "TestSimulationFreeSpheres","TestSimulationFreeSpheresShear",
           "TestSingleSphereLinear","TestSingleSphereRotation",
           "TestTwoSpheresCentral","TesttwoSpheresGeneral",
           "TestTwoSpheresLeesEdwards","TestTwoSpheresTangential"]



# Tests
from .test_oneSphereFreeOneSphereBoundary import TestoneSphereFreeOneSphereBoundary
from .test_simulationCrosslinkedShear import TestSimulationCrosslinkedShear
from .test_simulationFreeSpheres import TestSimulationFreeSpheres
from .test_simulationFreeSpheresShear import TestSimulationFreeSpheresShear
from .test_singleSphereLinear import TestSingleSphereLinear
from .test_singleSphereRotation import TestSingleSphereRotation
from .test_twoSpheresCentral import TestTwoSpheresCentral
from .test_twoSpheresGeneral import TesttwoSpheresGeneral
from .test_twoSpheresLeesEdwards import TestTwoSpheresLeesEdwards
from .test_twoSpheresTangential import TestTwoSpheresTangential











from tkinter import *
import random
import math
from .Ensemble import Ensemble
from particleShearBase import Force_register

class EnsembleFriction(Ensemble):
    """Ensemble of particles of type  `particleShear.SphereFriction`, with regular boundary conditions

            This class creates and maintains a list of objects of type
            `particleShear.SphereFriction`. These objects are generically
            referred to as spheres and are stored in `particleShear.CanvasPoints.sphereList`.\n
            The ensemble class does not implement proper Lees-Edwards boundaries, but emulates neighboring spheres by
            creating replicates of the spheres near the boundares displaced by one positive or negative multiple of the
            xy dimensions of the simulation area.\n
            Class defined in subpackage particleShearObjects"""

    def __init__(self, size_x, size_y, N, packing_fraction=0.8, theCanvas=FALSE, doDrawing=FALSE, k=1, nu=0.01,m=1,
                 bimodal_factor=1.4,
                 k_t=1,nu_t=0.01,mu=0.1):

        """Initialize self

                - **parameters**\n
                `size_x` Width of area to be used in pixels = micrometers for the simulation\n
                `size_y` Height of area to be used in pixels = micrometers for the simulation\n
                `N` The number of spheres to place
                `packing_fraction` The packing fraction indicating the density; this is defined as the area occupied by the
                non-compressed spheres as compared to the actual available area (size_x times size_y)
                `theCanvas` possibility to transmit a `tkinter` `Canvas` object for graphical output\n
                `doDrawing` Flag to indicate whether graphical output should be produced or not
                `k` Central spring constant in (mg/s^2)/m of depth;
                so F=-k*x*L with L the depth of the stack (the simulation is 2D) and x the compression\n
                `nu` Central viscosity force constant in (mg/s)/m of depth. This is to have F=nu*v*L, L again the depth\n
                `m` is the average mass per unit of depth in mg/m of depth\n
                `bimodal_factor` is ratio of radii of the smaller and larger spheres in the bimodal distribution\n
                `k_t` Tangential spring constant in (mg/s^2)/m of depth;
                relevant for locked or permanent interfaces but not for slipping interfaces\n
                `nu_t` Central viscosity force constant in (mg/s)/m of depth.
                 Relevant for frictionally locked and permanent interfaces but not slipping ones\n
                `mu` Friction coefficient, describes the maximum interface force for non-permanent interfaces by
                 F_tangential_max = F_central*mu
                """

        super(EnsembleFriction, self).__init__(size_x, size_y, N, packing_fraction, theCanvas, doDrawing, k, nu,m,
                                               bimodal_factor=bimodal_factor)

        self.N = N
        """Number of spheres"""
        self.packing_fraction = packing_fraction
        """Packing fraction: relative area occupied by the uncompressed spheres as compared to the available area"""

        self.k_t=k_t
        """Transversal spring constant in (mg/s^2)/m of depth"""

        self.nu_t=nu_t
        """Transversal viscosity constant in (mg/s)/m of depth"""

        self.mu=mu
        """Friction coefficient"""

    def particle_info(self):
        """Provide short description of particle type of inclusion into output file

        Method defined in `particleShear.EnsembleFriction` """
        return "Frictional spheres, \n\tbimodal distribution, equal probability, \n\tr1="+\
               str(self.r_max*0.8/1.2)+"r2="+str(self.r_max)+"\n\tidentical mass="+str(self.sphereList[0].m)+"mg"



    def tangential_force(self):
        """Let all spheres calculate the total tangental force acting on them.

                 This will check for contact and so is more time
                 consuming than `particleShear.EnsembleFriction.tangential_force_from_neighbors`.
                 For this function to work, the spheres must be of type `particleShear.SphereFriction`or a
                 subclass thereof, or otherwise possess a method `tangential_force`\n
                 This method is defined in class `particleShear.EnsembleFriction`"""


        for sphereIndex in range(len(self.sphereList)):
            for sphereIndex2 in range(len(self.sphereList)):
                if sphereIndex != sphereIndex2:
                    self.sphereList[sphereIndex].tangential_force(self.sphereList[sphereIndex2],
                                                                  self.nu_t,self.mu,self.k,self.k_t)
            for sphereIndexBoundary in range(len(self.spheres_boundary)):
                self.sphereList[sphereIndex].tangential_force(self.spheres_boundary[sphereIndexBoundary],
                                                              self.nu_t,self.mu,self.k,self.k_t)
    def tangential_force_from_neighbors(self):
        """Let all spheres calculate the total tangential force acting on from their known neighbors.

                This is quicker than `particleShear.EnsembleFriction.tangential_force`
                but necessitates the neighbor relations to be already established via
                `particleShear.CanvasPointsNeighbors.test_neighbor_relation`.
                For this function to work, the spheres must be of type `particleShear.SphereFriction`or a
                subclass thereof, or otherwise possess a method `tangential_force_from_neighbors`.\n
                This method is defined in class `particleShear.EnsembleFriction`"""

        for sphereIndex in range(len(self.sphereList)):
            self.sphereList[sphereIndex].tangential_force_from_neighbors(self.nu_t,self.mu,self.k,self.k_t)

    def do_rotational_acceleration(self, dt):
        """Have the spheres calculate their rotational acceleration from the tangential forces.

                For this function to work,
                the spheres must be of type `particleShear.SphereFriction` or subclass thereof, or alternatively,
                possess a method `do_rotational_acceleration`\n
                This method is defined in class `particleShear.EnsembleFriction`"""
        for theSphere in self.sphereList:
            theSphere.do_rotational_acceleration(dt)


    def mechanical_simulation_step_calculate_forces(self):
        """ Calculate the forces acting on the spheres

        This method is defined in class `particleShear.EnsembleFriction`"""
        super(EnsembleFriction,self).mechanical_simulation_step_calculate_forces()
        self.tangential_force_from_neighbors()


    def mechanical_simulation_step_calculate_acceleration(self,cool_factor=0.97,dt=1):
        """Calculate the accelerations resulting from the forces

        This method is defined in class `particleShear.EnsembleFriction`"""
        self.do_linear_acceleration(dt)
        self.do_rotational_acceleration(dt)
        self.cool(cool_factor)


























import unittest
from particleShear import *
from tkinter import *
import math



class TestTwoSpheresTangential(unittest.TestCase):

    def setUp(self):



        size_x = 500
        size_y = 500

        theTk = Tk()
        theCanvas = Canvas(theTk, width=size_x, height=size_y)

        theCanvas.pack()

        theEnsemble = EnsembleLinkable(500, 500, 0, 0.8, theCanvas, True, k=1, nu=1, k_t=1, nu_t=1, mu=0.01)

        theEnsemble.sphereList.append(SphereLinkable("GREY", 150, 200, 120, m=1, my_index=1, theCanvas=theCanvas,
                                                     doDrawing=True, force_register=theEnsemble,
                                                     size_x=theEnsemble.size_x,
                                                     size_y=theEnsemble.size_y))

        theEnsemble.sphereList[0].xspeed = 0

        theEnsemble.sphereList.append(SphereLinkable("BLUE", 220, 260, 120, m=1, my_index=2, theCanvas=theCanvas,
                                                     doDrawing=True, force_register=theEnsemble,
                                                     size_x=theEnsemble.size_x,
                                                     size_y=theEnsemble.size_y))

        theEnsemble.sphereList[1].xspeed = 0

        theEnsemble.sphereList[1].omega = 0.25

        graphical = theEnsemble.graphical_output_configuration

        graphical.draw_rotation_line = True

        graphical.draw_permanent_connections=True

        theEnsemble.set_graphical_output_configuration(graphical)

        theEnsemble.sphereList[0].establish_permanent_link(theEnsemble.sphereList[1])

        theTk.update()

        theEvaluator = StressTensorEvaluation(size_x, size_y)

        self.theEvaluator = theEvaluator
        self.theEnsemble = theEnsemble
        self.theTk=theTk
        self.theEnsemble.k = 0
        self.theEnsemble.nu = 0


        self.dt=0.01
        self.cool_factor=1

        for i in range(50):
            self.theEnsemble.mechanical_simulation_step_calculate_forces()

            self.theEnsemble.record_total_particle_forces()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)

            # Chosen evaluation step: Take the coordinates before moving the spheres, these coordinates serve for the
            # basis of force calculation. The inter-sphere vector is
            self.L = [self.theEnsemble.sphereList[1].x - self.theEnsemble.sphereList[0].x,
                      self.theEnsemble.sphereList[1].y - self.theEnsemble.sphereList[0].y]


            # Generally, due to the change of friction position, the force can vary. This is only an issue with
            # k_t
            self.Ft1=self.theEnsemble.force_register.pair_register[6][2]
            self.Ft2 = self.theEnsemble.force_register.pair_register[5][2]
            self.Ft=[(self.Ft1[0]+self.Ft2[0])/2,(self.Ft1[1]+self.Ft2[1])/2 ]


            self.Ft_abs = math.sqrt(self.Ft[0] * self.Ft[0] + self.Ft[1] * self.Ft[1])
            self.L_abs = math.sqrt(self.L[0] * self.L[0] + self.L[1] * self.L[1])

            self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
            self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
            self.theTk.update()


    def tearDown(self):
        self.theTk.destroy()
        self.theTk = False





    def test_stress_tensor_LW(self):





        self.assertTrue(abs((self.Ft[0]*self.L[0]+self.Ft[1]*self.L[1])/self.L_abs/self.Ft_abs)<1e-15)

        test_tensor = [[0, 0], [0, 0]]

        # The idea here is that the reported transversal forces in the pair register are half of the force
        # because they are split between the two pairs

        test_tensor[0][0] = self.Ft[0] * self.L[0] * 1e-18 * 2 / (self.theEnsemble.size_x * 1e-6) / (
                    self.theEnsemble.size_y * 1e-6)
        test_tensor[0][1] = self.Ft[0] * self.L[1] * 1e-18 * 2 / (self.theEnsemble.size_x * 1e-6) / (
                self.theEnsemble.size_y * 1e-6)
        test_tensor[1][0] = self.Ft[1] * self.L[0] * 1e-18 * 2 / (self.theEnsemble.size_x * 1e-6) / (
                self.theEnsemble.size_y * 1e-6)
        test_tensor[1][1] = self.Ft[1] * self.L[1] * 1e-18 * 2 / (self.theEnsemble.size_x * 1e-6) / (
                self.theEnsemble.size_y * 1e-6)

        LW = self.theEvaluator.stress_tensor_LW





        self.assertTrue(abs(LW[0][0])>0 or abs(LW[0][1])>0 or abs(LW[1][0])>0 or abs(LW[1][1])>0)

        self.assertTrue(abs((LW[0][0]-test_tensor[0][0])/test_tensor[0][0])<1e-15 and
                        abs((LW[1][0] - test_tensor[1][0]) / test_tensor[1][0]) < 1e-15 and
                        abs((LW[0][1] - test_tensor[0][1]) / test_tensor[0][1]) < 1e-15 and
                        abs((LW[1][1] - test_tensor[1][1]) / test_tensor[1][1]) < 1e-15)

    def test_overall_stress_tensor(self):



        overall = self.theEvaluator.overall_stress_tensor

        overall_tot = abs(overall[0][0])+abs(overall[0][1])+abs(overall[1][0])+abs(overall[1][1])

        self.assertTrue(abs((overall[0][1] - overall[1][0]) / overall_tot) < 1e-15)

    def test_unbalanced_forces_stress_tensor(self):




        unbalanced = self.theEvaluator.stress_tensor_unbalanced_forces

        LW = self.theEvaluator.stress_tensor_LW


        LW_tot=abs(LW[0][0])+abs(LW[0][1])+abs(LW[1][0])+abs(LW[1][1])

        self.assertTrue(abs(LW[0][0] - unbalanced[0][0])/LW_tot < 1e-15 and
                        abs(LW[0][1] - unbalanced[0][1]) / LW_tot < 1e-15 and
                        abs(LW[1][0] - unbalanced[1][0]) / LW_tot < 1e-15 and
                        abs(LW[1][1] - unbalanced[1][1]) / LW_tot < 1e-15
                        )

    def test_external_stress_tensor(self):


        external = self.theEvaluator.stress_tensor_with_external_forces
        self.assertTrue(
            external[0][0] == 0 and external[0][1] == 0 and external[1][0] == 0 and external[1][1] == 0)

    def test_peculiar_acceleration_stress_tensor(self):


        peculiar = self.theEvaluator.stress_tensor_linear_acceleration_otsuki

    def test_stress_tensor_spin_kinetic_energy(self):

        stress_tensor_spin = self.theEvaluator.stress_tensor_spin_kinetic_energy

        # In theory, the spin kinetic energy should be I omega*w^2

        self.assertTrue(
            stress_tensor_spin[0][1] == stress_tensor_spin[1][0] )

    def test_stress_tensor_unbalanced_torque(self):

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)

        stress_tensor_torque = self.theEvaluator.stress_tensor_unbalanced_torque

        self.assertTrue(
             stress_tensor_torque[0][1] == -stress_tensor_torque[1][0])


    def test_conservation_linear_momentum(self):

        self.theEnsemble.mechanical_simulation_step_calculate_forces()

        self.theEnsemble.record_total_particle_forces()
        self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)
        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
        self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
        self.theTk.update()

        old_linear_momentum = theHandler.total_mass * theHandler.v_mean

        for i in range(100):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()

        self.theEnsemble.mechanical_simulation_step_calculate_forces()

        self.theEnsemble.record_total_particle_forces()
        self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)
        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
        self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
        self.theTk.update()

        new_linear_momentum = theHandler.total_mass * theHandler.v_mean

        if(math.sqrt(new_linear_momentum[0]*new_linear_momentum[0]+\
                                        new_linear_momentum[1]*new_linear_momentum[1])>1e-5):

            self.assertTrue(abs(old_linear_momentum[0] - new_linear_momentum[0]) \
                            / math.sqrt(new_linear_momentum[0]*new_linear_momentum[0]+\
                                        new_linear_momentum[1]*new_linear_momentum[1]) < 1e-15)

            self.assertTrue(abs(old_linear_momentum[1] - new_linear_momentum[1]) \
                            / math.sqrt(new_linear_momentum[0]*new_linear_momentum[0]+\
                                        new_linear_momentum[1]*new_linear_momentum[1]) < 1e-15)

        else:
            self.assertTrue(abs(old_linear_momentum[0]-new_linear_momentum[0])<1e-15)
            self.assertTrue(abs(old_linear_momentum[1] - new_linear_momentum[1]) < 1e-15)

    def test_conservation_angular_momentum(self):

        self.theEnsemble.mechanical_simulation_step_calculate_forces()

        self.theEnsemble.record_total_particle_forces()
        self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)
        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
        self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
        self.theTk.update()

        old_angular_momentum = theHandler.angular_momentum_around_origin

        for i in range(100):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()

        self.theEnsemble.mechanical_simulation_step_calculate_forces()

        self.theEnsemble.record_total_particle_forces()
        self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)
        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
        self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
        self.theTk.update()

        new_angular_momentum = theHandler.angular_momentum_around_origin


        self.assertTrue(abs((old_angular_momentum - new_angular_momentum) / new_angular_momentum) < 1e-14)


if __name__ == '__main__':
    unittest.main()



from particleShearBase import Force_register

from particleShearBase import CircleBasicElasticity

# Basic friction-less sphere with drawing functions
class Sphere(CircleBasicElasticity):
    """Objects of type `particleShear.CircleBasicElasticity`, using standard rather than periodic Lees-Edwards
        boundary conditions"""
    def __init__(self, color, x, y, diameter, m=1,my_index=1,theCanvas=False, doDrawing=False,force_register=Force_register()):
        """Constructor as `particleShearBase.CircleBasicElasticity.__init__` but with use_lees_edwards set to False
        and indexing element (my_index)."""

        super(Sphere,self).__init__(color,x,y,diameter,m=m, theCanvas=theCanvas, doDrawing=doDrawing,force_register=force_register,
                                    use_lees_edwards=False)

        self.myindex=my_index # To give a unique index









from .EvaluationHandler import EvaluationHandler

class EvaluationHandlerPlotter(EvaluationHandler):
    def __init__(self, theTk, theCanvas, x_scale=1, y_scale_strain=1, y_scale_force=1, r=10, offset_x=0, offset_y=0,
                 plotStress=True):
        self.plotStress=plotStress
        super(EvaluationHandlerPlotter, self).__init__()
        self.theTk = theTk
        self.theCanvas = theCanvas
        self.x_scale = x_scale
        self.y_scale_strain = y_scale_strain
        self.y_scale_force = y_scale_force
        self.r = r
        self.offset_x = offset_x
        self.offset_y = offset_y



    def record(self, theEnsemble,theEvaluator=False):
        super(EvaluationHandlerPlotter, self).record(theEnsemble,theEvaluator)

        if self.plotStress:
            t = self.t[len(self.t) - 1]
            strain = self.strain[len(self.strain) - 1]
            force = self.force[len(self.force) - 1]
            shear_stress = self.shear_stress[len(self.shear_stress) - 1]

            self.plot(t, strain, force - self.force[0], shear_stress - self.shear_stress[0])



    def plot(self, t, strain, force,shear_stress):
        # First, x coordinates (common)



        x1 = t * self.x_scale - self.r + self.offset_x
        x2 = x1 + 2 * self.r



        # Then, y for volume shear stress
        y1 = -shear_stress * self.y_scale_force - self.r + self.offset_y #Here, the minus sign is just because of the inverted
        # coordinate system of the canvas graphics (tkinter)
        y2 = y1 + 2 * self.r
        self.theCanvas.create_oval(x1, y1, x2, y2, fill="red",outline="red")

        # Then, y for force
        y1 = -force * self.y_scale_force - self.r + self.offset_y  # Here, the minus sign is just because of the inverted
        # coordinate system of the canvas graphics (tkinter)
        y2 = y1 + 2 * self.r
        self.theCanvas.create_oval(x1, y1, x2, y2, fill="blue", outline="blue")

        # Then, y for strain
        y1 = -strain * self.y_scale_strain - self.r + self.offset_y
        y2 = y1 + 2 * self.r
        self.theCanvas.create_oval(x1, y1, x2, y2, fill="green",outline="green")
        self.theTk.update()



from tkinter import *
import random
import math

from particleShearBase import CanvasPointsFrictionElasticityLeesEdwards

def dotProduct2(a,b):
    """Helper function for the scalar product of vectors of length 2"""
    return a[0]*b[0]+a[1]*b[1]

class CanvasPointsLinkable (CanvasPointsFrictionElasticityLeesEdwards):
    """Canvas for placing `particleShear.SphereLinkable` objects (referred to as spheres).
          This class maintains a list of objects of type
          `particleShear.SphereLinkable` and adds elementary functions related to crosslinking"""

    def __init__(self, size_x, size_y, theCanvas=FALSE, doDrawing=False,
                 k=1, nu=0.01,m=1,k_t=1,nu_t=0.01,mu=0.1):

        """Initialize self

                                - **parameters**\n
                                    `size_x` Width of area to be used in pixels = micrometers for the simulation\n
                                    `size_y` Height of area to be used in pixels = micrometers for the simulation\n
                                    `theCanvas` possibility to transmit a `tkinter` `Canvas` object for graphical output\n
                                    `doDrawing` Flag to indicate whether graphical output should be produced or not\n
                                    `k` Central spring constant in (mg/s^2)/m of depth;
                                    so F=-k*x*L with L the depth of the stack (the simulation is 2D) and x the compression\n
                                    `nu` Central viscosity force constant in (mg/s)/m of depth. This is to have F=nu*v*L, L again the depth\n
                                    `m` is the mass per unit of depth in mg/m of depth\n
                                    `k_t` Tangential spring constant in (mg/s^2)/m of depth;
                                    relevant for locked or permanent interfaces but not for slipping interfaces\n
                                    `nu_t` Central viscosity force constant in (mg/s)/m of depth.
                                     Relevant for frictionally locked and permanent interfaces but not slipping ones\n
                                    `mu` Friction coefficient, describes the maximum interface force for non-permanent interfaces by
                                    F_tangential_max = F_central*mu"""

        super(CanvasPointsLinkable, self).__init__(
            size_x, size_y, theCanvas=theCanvas, doDrawing=doDrawing,
                 k=k, nu=nu,m=m,k_t=k_t,nu_t=nu_t,mu=mu)


    def particle_info(self):
        """Provide short description of particle type of inclusion into output file

                Method defined in `particleShear.CanvasPointsLinkable` """
        return "CanvasPointsLinkable: Frictional spheres with possible permanent links, Lees Edwards boundary"


    def makeAllLinksPermanent(self):
        """Link all currently touching spheres with permanent links

                        Method defined in `particleShear.CanvasPointsLinkable` """
        for theSphere in self.sphereList:
            for theNeighbor in theSphere.neighbors:
                theSphere.establish_permanent_link(theNeighbor.theSphere)

    def cutLine(self,point,angle):
        """Remove permanent links across a straight line

        - **parameters**\n
            `point` Vector of 2 elements for x and y coordinates of a point on the cut line\n
            `angle` Angle of the cut line, in radians; 0=vertical \n
            Method defined in `particleShear.CanvasPointsLinkable` """
        norm_vector = [math.cos(angle), math.sin(angle)]
        base = dotProduct2(norm_vector,point)

        for theSphere in self.sphereList:
            for theNeighbor in theSphere.neighbors:
                if (theNeighbor.interface_type == "permanent"):
                    myScalar_product = dotProduct2(norm_vector, [theSphere.x, theSphere.y]) - base
                    dist = theSphere.d(theNeighbor.theSphere)
                    n_vector = theSphere.n(theNeighbor.theSphere)
                    neighbor_scalar_product = dotProduct2(norm_vector, [theSphere.x + n_vector[0] * dist,
                                                                      theSphere.y + n_vector[1] * dist]) - base
                    # Crossline means opposite side or cut through one or two of the points
                    if (myScalar_product * neighbor_scalar_product) <= 0:
                        theSphere.cut_permanent_link(theNeighbor.theSphere)

    def cutRandomLine(self):
        """Remove permanent links across a random cut line

            Method defined in `particleShear.CanvasPointsLinkable` """
        point=[random.randint(0,self.size_x), random.randint(0,self.size_y)]
        angle=random.random()* 2 * math.pi
        self.cutLine(point,angle)

    def cutTopBottomEdge(self):
        """Remove permanent links across a top and bottom boundary lines

            Method defined in `particleShear.CanvasPointsLinkable` """
        self.cutLine([0,0],math.pi/2)
        self.cutLine([0, self.size_y], math.pi / 2)










from tkinter import *
import random
import math

from .Force_register import Force_register

from .neighbor_relation import neighbor_relation

from .Graphical_output_configuration import Graphical_output_configuration

from .CircleMass import CircleMass

# Basic friction-less sphere with drawing functions
class CircleMassNeighbors(CircleMass):
    """Provide a circle with a neighbor list

       Add capacity to handle the relation to neighboring spheres to the `particleShear.CircleMass` parent class """
    def __init__(self, color, x, y, diameter, m=1, theCanvas=False, doDrawing=False,force_register=Force_register(),
                 use_lees_edwards=False):



        self.neighbors = []
        """The list of neighbors, of class `particleShear.neighbor_relation` (or subclass)"""

        super(CircleMassNeighbors,self).__init__(color,x,y,diameter,m=m, theCanvas=theCanvas, doDrawing=doDrawing,
                                                 force_register=force_register,use_lees_edwards=use_lees_edwards)







    def set_graphical_output_configuration(self, graphical_output_configuration):
        """Set new `particleShear.Graphical_output_configuration`

        This method is defined in class `particleShear.CircleMassNeighbors`"""
        self.deleteNeighborDrawing()
        super(CircleMassNeighbors,self).set_graphical_output_configuration(graphical_output_configuration)
        self.initiateNeighborDrawing()



    def deleteNeighborDrawing(self):
        """Deletes the drawing objects for the neighbor interfaces

        There seems to be no good way to move a line, so deleting and drawing
        is done at every simulation step for the neighbor lines.\n
        This method is defined in class `particleShear.CircleMassNeighbors`"""
        for theNeighbor in self.neighbors:
            if theNeighbor.graphical_line_index!=-1:
                if self.theCanvas:
                    self.theCanvas.delete(theNeighbor.graphical_line_index)
                theNeighbor.graphical_line_index=-1

    def contactLineColor(self,interface_type):
        """Find current contact line color

        Color codes for the interface state; by default, red is locked, green is slipping.
        These settings can be changed in the `particleShear.CircleMassNeighbors.graphical_output_configuration` instance
        variable\n
        This method is defined in class `particleShear.CircleMassNeighbors`"""
        col=self.graphical_output_configuration.color_interface_slip
        if(interface_type == "stick"):
            col=self.graphical_output_configuration.color_interface_locked
        return col


    def initiateNeighborDrawing(self):
        """Generate the graphical representation of the neighboring lines

        This method is defined in class `particleShear.CircleMassNeighbors`"""



        if self.graphical_output_configuration.draw_active_interface:
            for theNeighbor in self.neighbors:
                if not theNeighbor.interface_type=="permanent" or self.graphical_output_configuration.draw_permanent_interfaces:
                    if theNeighbor.graphical_line_index == -1:
                        dist = self.d(theNeighbor.theSphere)
                        # From geometrical considerations: intersection circles with shared height segment
                        # self.r^2-x^2=theNeighbor.theSphere.r^2-(dist-x)^2
                        # self.r^2-x^2+dist^2-2*x*dist+x^2=theNeighbor.theSphere.r^2
                        # self.r^2+dist^2-2*x*dist=theNeighbor.theSphere.r^2
                        # dist^2-2*x*dist=theNeighbor.theSphere.r^2-self.r^2
                        # -dist^2+2*x*dist=self.r^2-theNeighbor.theSphere.r^2
                        # -dist+2*x=(self.r^2-theNeighbor.theSphere.r^2)/dist
                        # x=1/2*(dist+(self.r^2-theNeighbor.theSphere.r^2)/dist)

                        if (dist <= self.r + theNeighbor.theSphere.r) and dist > 0:
                            l = (dist + (
                                        self.r * self.r - theNeighbor.theSphere.r * theNeighbor.theSphere.r) / dist) / 2

                            if self.r * self.r - l * l >= 0:
                                h = math.sqrt(self.r * self.r - l * l)
                                norm_vector = self.n(theNeighbor.theSphere)
                                pos = self.coordinates()
                                touch_point_x = pos[0] + norm_vector[0] * l
                                touch_point_y = pos[1] + norm_vector[1] * l
                                endpoint_1_x = touch_point_x - norm_vector[1] * h
                                endpoint_1_y = touch_point_y + norm_vector[0] * h
                                endpoint_2_x = touch_point_x + norm_vector[1] * h
                                endpoint_2_y = touch_point_y - norm_vector[0] * h
                                color = self.contactLineColor(theNeighbor.interface_type)
                                if self.theCanvas:
                                    theNeighbor.graphical_line_index = self.theCanvas.create_line(
                                        endpoint_1_x,
                                        endpoint_1_y,
                                        endpoint_2_x,
                                        endpoint_2_y,
                                        width=3, fill=color
                                    )

    def moveDrawing(self, dt):
        """Move the drawings associated with this object

        This method is defined in class `particleShear.CircleMassNeighbors`"""
        super(CircleMassNeighbors,self).moveDrawing(dt)
        self.deleteNeighborDrawing()
        self.initiateNeighborDrawing()

    def deleteDrawing(self):
        """Delete the drawings associated with this object

        This method is defined in class `particleShear.CircleMassNeighbors`"""
        self.deleteNeighborDrawing()
        super(CircleMassNeighbors,self).deleteDrawing()



    def test_neighbor_relation(self, theSphere):
        """Test whether a given sphere is a neighbor

        Tests whether or not the sphere is a geometrical neighbor. Depending on the result, update the list of
        `particleShear.CircleMassNeighbors.neighbors`. If necessary, draw or delete the associated shapes.\n
        This method is defined in class `particleShear.CircleMassNeighbors`"""
        if abs(self.y-theSphere.y) > self.r+theSphere.r and abs(self.y-theSphere.y)< self.size_y-self.r-theSphere.r:
            is_neighbor=False
        else:
            d = self.d(theSphere)
            pos = theSphere.coordinates()
            r = pos[2]
            is_neighbor = (d < r + self.r)


        # Regardless of the distance, first check whether we find the sphere in the neighbors list
        found = FALSE
        foundNeighborIndex=-1
        for theNeighborIndex in range(len(self.neighbors)):
            theNeighbor = self.neighbors[theNeighborIndex]
            if (theSphere == theNeighbor.theSphere):
                foundNeighborIndex = theNeighborIndex
                found = TRUE

        if not is_neighbor: #If the sphere is not a neighbor (anymore)
            if found: #It was previously a neighbor, but no more, so delete it from the neighbors list
                if self.doDrawing:
                    self.deleteNeighborDrawing()
                del self.neighbors[foundNeighborIndex]
                if self.doDrawing:
                    self.initiateNeighborDrawing()
        if is_neighbor:
            if not found: # Not found previously, so add a new entry
                info = neighbor_relation(0,"stick",theSphere)
                if self.doDrawing:
                    self.deleteNeighborDrawing()
                self.neighbors.append(info)
                if self.doDrawing:
                    self.initiateNeighborDrawing()



    def findNeighborIndex(self, theSphere):
        """Return whether theSphere is among the currently listed `particleShear.CircleMassNeighbors.neighbors`"""

        foundNeighborIndex = -1
        for theNeighborIndex in range(len(self.neighbors)):
            theNeighbor = self.neighbors[theNeighborIndex]
            if (theSphere == theNeighbor.theSphere):
                foundNeighborIndex = theNeighborIndex

        return foundNeighborIndex


import math

class Force_register():
    """Provide accounting of the forces acting on the particles.

    Detailed accounting of the forces acting on the particles allows to calculate the relevant
    average stress tensors at each simulation time point by means of the
     `particleShear.StressTensorEvaluation` class.\n
     Sub-package particleShearBase"""

    def __init__(self):
        """Initialize the force register
        Provide default empty intance variables"""
        self.pair_register=[]
        """ Hold the elementary internal pairwise forces"""
        self.internal_moment_register = []
        """ Hold the fully internally generated torques"""
        self.total_particle_force_register=[]
        """ Hold the single total forces acting single particles"""
        self.external_force_register=[]
        """Hold single forces acting from the outside (across the canvas boudnary) on the particles"""
        self.external_moment_register = []
        """Hold single torques acting from the outside (across the canvas boudnary) on the particles"""
        self.unbalanced_moment_register = []
        """Hold the net torque moments causing rotational acceleration"""

    def reset_force_register(self):
        """Reset the force register to its original state

        This method is defined in class `particleShear.Force_register`"""

        self.pair_register = []

        self.total_particle_force_register = []  #

        self.external_force_register = []

        self.unbalanced_moment_register=[]

        self.internal_moment_register=[]

        self.external_moment_register = []

    def record_individual_internal_force(self,target,source,force_vector):
        """Record an individual internal force.

         The force magnitude and direction is described by the force_vector. The target describes the sphere on
         which the force acts (it is applied to the center); the source describes the sphere from which the force
         originates. Due to Newton's law of action an reaction, there should always be a second entry in the
         `particleShear.Force_register.pair_register` with opposing force vector, and exchanged source and target. However,
         this function does not check for the presence of such a homologous entry, this needs to be done in the code
         invoking this method.\n
         This method is defined in class `particleShear.Force_register`"""
        self.pair_register.append([target,source,force_vector])

    def record_total_particle_force(self,target,force_vector):
        """Record a net total particle force

        As there should be only one net particle force for each particle, the method first looks
        for an entry corresponding to target. If such an entry is found, the force is updated, otherwise a new
        entry is created

         Method defined in class `particleShear.Force_register`"""

        # First, look whether there is already an entry

        found_index = -1

        for ind in range(len(self.total_particle_force_register)):
            if self.total_particle_force_register[ind][0]==target:
                found_index=ind

        # If there is an entry, update it
        if(found_index>=0):
            self.total_particle_force_register[found_index][1]=force_vector
        else: # else add one
            self.total_particle_force_register.append([target,force_vector])

    def record_external_force(self, target, force_vector):
        """ Add an external force

        External forces are coming from either outside the area under examination or from that is not free to move
         (see `particleShear.CanvasPointsShear.canMove`)\n
         Method defined in class `particleShear.Force_register`"""

        self.external_force_register.append ([target, force_vector])

    def record_external_torque(self, target, moment):
        """ Add an external torque

        External torques are coming from either outside the area under examination or from that is not free to move
         (see `particleShear.CanvasPointsShear.canMove`)\n
         Method defined in class `particleShear.Force_register`"""

        self.external_moment_register.append ([target, moment])


    def record_individual_internal_torque(self,target,source,moment):
        """Record an individual internal moment.

         The torque is described by the moment argument. The target describes the sphere on
         which the torque acts (it is applied homogeneously to cause correct acceleration);
         the source describes the sphere which by frictional interaction has helped
         generate the moment (typically, in `particleShear.CircleFrictionElasticity.distribute_tangential_couple`).
         Since frictional interaction tends to generate
          torque on both partners, there is generally a second entry with target and source inversed. However,
         this function does not check for the presence of such a homologous entry, this needs to be done in the code
         invoking this method.\n
         This method is defined in class `particleShear.Force_register`"""
        self.internal_moment_register.append([target,source,moment])


    def record_unbalanced_moment(self,target,moment):
        """ Add net torque on a sphere

        As there should be only one net particle torque for each particle, the method first looks
        for an entry corresponding to target. If such an entry is found, the force is updated, otherwise a new
        entry is created

        Method defined in class `particleShear.Force_register`"""

        found_index = -1

        for ind in range(len(self.unbalanced_moment_register)):
            if self.unbalanced_moment_register[ind][0] == target:
                found_index = ind

        # If there is an entry, update it
        if (found_index >= 0):
            self.unbalanced_moment_register[found_index][1] = moment
        else:  # else add one
            self.unbalanced_moment_register.append([target, moment])








"""Provide 2D geometrical fricional and non-frictional objects (circles) and canvas areas to place, control and
 display them.

The package provides the tools necessary to simulate ensembles of discrete spherical particles as described in
Otsuki, M. and H. Hayakawa, Discontinuous change of
shear modulus for frictional jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902.\n
Module written by Thomas Braschler (thomas.braschler@gmail.com)
"""

__all__ = ["Force_register","StressTensorEvaluation","Graphical_output_configuration","neighbor_relation",
           "Point","PointLeesEdwards",
           "Circle","CircleBasicElasticity",
           "CircleFrictionElasticity","CircleMass","CircleMassNeighbors",
           "CanvasPoints","CanvasPointsBasicElasticity","CanvasPointsBasicElasticityLeesEdwards",
           "CanvasPointsFrictionElasticityLeesEdwards","CanvasPointsMass","CanvasPointsShear","CanvasPointsNeighbors",
           "elastic_force_law_plateau","elastic_force_law","PlateauConfiguration","distance_transform_plateau"]



# Basic tools
from .Force_register import Force_register
from .StressTensorEvaluation import StressTensorEvaluation
from .Graphical_output_configuration import Graphical_output_configuration
from .neighbor_relation import neighbor_relation



# Geometrical elements
from .Point import Point
from .PointLeesEdwards import PointLeesEdwards
from .Circle import Circle
from .CircleBasicElasticity import CircleBasicElasticity
from .CircleFrictionElasticity import CircleFrictionElasticity
from .CircleMass import CircleMass
from .CircleMassNeighbors import CircleMassNeighbors


# Canvases for placing the elements

from .CanvasPoints import CanvasPoints
from .CanvasPointsBasicElasticity import CanvasPointsBasicElasticity
from .CanvasPointsBasicElasticityLeesEdwards import CanvasPointsBasicElasticityLeesEdwards
from .CanvasPointsFrictionElasticityLeesEdwards import CanvasPointsFrictionElasticityLeesEdwards
from .CanvasPointsMass import CanvasPointsMass
from .CanvasPointsNeighbors import CanvasPointsNeighbors
from .CanvasPointsShear import CanvasPointsShear

# Callback functions

from .CircleBasicElasticity import elastic_force_law
from .CircleBasicElasticity import elastic_force_law_plateau
from .CircleBasicElasticity import PlateauConfiguration
from .CircleBasicElasticity import distance_transform_plateau









from tkinter import *
import random
import math

from particleShearBase import neighbor_relation
from particleShearBase import CircleBasicElasticity

from particleShearBase import Force_register

# Recognition across period boundary conditions with shear displacement
class SphereLeesEdwards(CircleBasicElasticity):
    """Objects of type `particleShear.CircleBasicElasticity`, using periodic Lees-Edwards
            boundary conditions"""
    def __init__(self, color, x, y, diameter, m=1, myindex=1,theCanvas=FALSE, doDrawing=FALSE,
                 force_register=Force_register(),size_x=500,size_y=500):
        """Constructor as `particleShearBase.CircleBasicElasticity.__init__` but with use_lees_edwards set to True
               and indexing element (my_index)."""


        super(SphereLeesEdwards, self).__init__(color, x, y, diameter, m, theCanvas, doDrawing,force_register,
                                                use_lees_edwards=True)
        self.size_x=size_x
        self.size_y=size_y
        self.myindex = my_index









from tkinter import *
import random
import math


# Basic moving point
class Point:
    """Define a moving point, with position (x and y) and speed (xpeed and yspeed)"""
    def __init__(self, x, y):
        """Initialize self

        - **parameters**\n
        `x` The x-position of the point (assumed micrometers for the simulation)\n
        `y` The y-position of the point (assumed micrometers for the simulation)\n
        """

        self.x = x
        """The x-position of the point (assumed micrometers for the simulation)"""
        self.y = y
        """The y-position of the point (assumed micrometers for the simulation)"""
        self.xspeed=0
        """The horizontal speed of the point (assumed micrometers/s for the simulation)"""
        self.yspeed=0
        """The vertical speed of the point (assumed micrometers/s for the simulation)"""


    def coordinates(self):
        """Return the current coordinates: vector of three elements, which are (x,y,radius=0)

        This method is defined in class `particleShearBase.Point`"""
        return [self.x, self.y, 0]

    def boundary_conditions(self, boundary_x, boundary_y):
        """Ensure location of the point within boundary conditions

        x should be >= 0 and < boundary_x\n
        y should be >= 0 and < boundary_y\n
        x will be adusted to satisfy this via addition or subtraction of multiples of boundary_x\n
        y will be adusted to satisfy this via addition or subtraction of multiples of boundary_y\n
        This method is defined in class `particleShearBase.Point`"""
        old_x = self.x
        old_y = self.y
        while self.x >= boundary_x:
            self.x = self.x - boundary_x
        while self.x < 0:
            self.x = self.x + boundary_x
        while self.y >= boundary_y:
            self.y = self.y - boundary_y
        while self.y < 0:
            self.y = self.y + boundary_y
        if self.doDrawing:
            self.theCanvas.move(self.shape, self.x - old_x, self.y - old_y)



    def d_euclidian(self,theSphere):
        """Calculate euclidian distance to another object of type `particleShearBase.Point`or derived

        The calculation is done via the Pythagorean formula: square root of the sum of the squares of the distances
        in x and y.\n
        This method is defined in class `particleShearBase.Point`"""
        pos = theSphere.coordinates()
        x = pos[0]
        y = pos[1]
        return math.sqrt((x - self.x) * (x - self.x) + (y - self.y) * (y - self.y))


    def d(self,theSphere):
        """Calculate the distance to another sphere.

        For this class, uses `particleShearBase.Point.d_euclidian` but is overriden by specific subclasses.\n
        This method is defined in class `particleShearBase.Point`"""
        return self.d_euclidian(theSphere)




    def n(self, theSphere):
        """Return unit vector pointing towards another sphere

        For this `particleShearBase.Point`, returns vector in cartesian coordinates pointing towards the center of the sphere
        indicated by the parameter theSphere. This behaviour can be overriden directly or indirectly by overriding
        the definition of `particleShearBase.Point.d`\n
        This method is defined in class `particleShearBase.Point`"""
        d = self.d(theSphere)
        if d > 0:
            pos = theSphere.coordinates()
            x = pos[0]
            y = pos[1]
            return ([(x - self.x) / d, (y - self.y) / d])
        if d == 0:
            angle = random.random()* 2 * math.pi
            return [math.cos(angle), math.sin(angle)]

    def relative_position(self,theSphere):
        """Return vector indicating the position of theSphere relative to myself

        For this `particleShearBase.Point`, returns vectorial difference between the position of theSphere and the current object,
        but this behaviour can be overriden directly, or also indirectly  when either the definition of
        `particleShearBase.Point.d` or `particleShearBase.Point.n` is overriden.\n
        This method is defined in class `particleShearBase.Point`"""
        n_vector = self.n(theSphere)
        d = self.d(theSphere)
        return [n_vector[0]*d,n_vector[1]*d]

    


    def move(self, dt=1):
        """Move for time interval dt

        This method is defined in class `particleShearBase.Point`"""
        self.x = self.x + self.xspeed * dt
        self.y = self.y + self.yspeed * dt




    def cool(self, f=0.8):
        """Cool by decreasing speed

        This method is defined in class `particleShearBase.Point`"""
        self.xspeed = self.xspeed * f
        self.yspeed = self.yspeed * f

    def relative_speed(self, theSphere):
        """Relative speed of the two spheres evaluated by the difference of speed of their centers

                This method is defined in class `particleShearBase.Point`"""
        return [theSphere.xspeed-self.xspeed, theSphere.yspeed-self.yspeed]



from tkinter import *
import random
import math
from .SphereLeesEdwards import SphereLeesEdwards
from particleShearBase import CanvasPointsBasicElasticityLeesEdwards
from .Ensemble import r_estimate
from .Ensemble import adjust_m
from particleShearBase import Force_register

class EnsembleLeesEdwards(CanvasPointsBasicElasticityLeesEdwards):
    """Ensemble of particles of type  `particleShear.SphereLeesEdwards`, with Lees-Edwards boundary conditions

                    This class creates and maintains a list of objects of type
                    `particleShear.SphereLeesEdwards`. These objects are generically
                    referred to as spheres and are stored in `particleShear.CanvasPoints.sphereList`.\n
                    This class uses Lees-Edwards boundary conditions\n
                    Class defined in subpackage particleShearObjects"""

    def __init__(self, size_x, size_y, N, packing_fraction=0.8, theCanvas=False, doDrawing=False, k=1, nu=0.01,m=1,
                 bimodal_factor=1.4):

        """Initialize self

                                - **parameters**\n
                                `size_x` Width of area to be used in pixels = micrometers for the simulation\n
                                `size_y` Height of area to be used in pixels = micrometers for the simulation\n
                                `N` The number of spheres to place
                                `packing_fraction` The packing fraction indicating the density; this is defined as the area occupied by the
                                non-compressed spheres as compared to the actual available area (size_x times size_y)
                                `theCanvas` possibility to transmit a `tkinter` `Canvas` object for graphical output\n
                                `doDrawing` Flag to indicate whether graphical output should be produced or not\n
                                `k` Central spring constant in (mg/s^2)/m of depth;
                                so F=-k*x*L with L the depth of the stack (the simulation is 2D) and x the compression\n
                                `nu` Central viscosity force constant in (mg/s)/m of depth. This is to have F=nu*v*L, L again the depth\n
                                `m` is the average mass per unit of depth in mg/m of depth\n
                                `bimodal_factor` is ratio of radii of the smaller and larger spheres in the bimodal distribution
                                """

        super(EnsembleLeesEdwards,self).__init__(size_x, size_y,  theCanvas=theCanvas, doDrawing=False, k=k, nu=nu,m=m)

        self.bimodal_upper = math.sqrt(bimodal_factor)
        """Upper multiplication factor for the bimodal sphere size distribution"""
        self.bimodal_lower = math.sqrt(1 / bimodal_factor)
        """Lower multiplication factor for the bimodal sphere size distribution"""

        self.N = N
        """Number of spheres"""
        self.packing_fraction = packing_fraction
        """Packing fraction: relative area occupied by the uncompressed spheres as compared to the available area"""

        self.doDrawing=doDrawing
        """Boolean: Should we draw the ensemble during simulation"""

        self.sphereList = []
        """List of the spheres"""

        r=self.size_y/2 # default value
        if N>0:
            r = r_estimate(self.size_x, self.size_y, packing_fraction, N,
                           bimodal_upper=self.bimodal_upper,
                           bimodal_lower=self.bimodal_lower)

        self.r_max = 0
        """Largest sphere radius"""
        for i in range(N):
            r_sphere = self.random_radius_bimodal(r)
            if r_sphere > self.r_max:
                self.r_max = r_sphere
            self.sphereList.append(SphereLeesEdwards("grey",
                                                     random.randrange(0, size_x),
                                                     random.randrange(0, size_y), r_sphere * 2,
                                                     adjust_m(m, r_sphere, r),
                                                     i, theCanvas, False,self,self.size_x,self.size_y))

        self.setShearInSpheres()

        self.initiateGraphics()


    def particle_info(self):
        """Provide short description of particle type of inclusion into output file

        Method defined in `particleShear.EnsembleLeesEdwards` """
        return "Non-frictional spheres with Lees-Edwards boundary conditions, \n\tbimodal distribution, equal probability, \n\tr1="+\
               str(self.r_max*0.8/1.2)+"r2="+str(self.r_max)+"\n\tidentical mass="+str(self.sphereList[0].m)+"mg"


    def random_radius_bimodal(self,r):
        """Choose randomly one of the two bimodal radius values.

        Method defined in `particleShear.EnsembleLeesEdwards`"""
        return r*(self.bimodal_lower+(self.bimodal_upper-self.bimodal_lower)*random.randint(0, 1))




# Particular particle definitions


__all__ = [
            "OscillatoryShearExperiment",
"OscillatorySimulation",
"OscillatorySimulationFragmentation","Simulation_dermal_filler_rheology",
"Simulation_interlocking_rheology",
"EvaluationHandler",
"EvaluationHandlerPlotter"]








# Simulation stuff
from .OscillatoryShearExperiment import OscillatoryShearExperiment # just the shear experiment
from .OscillatorySimulation import OscillatorySimulation # shear experiment and file saving
from .OscillatorySimulationFragmentation import OscillatorySimulationFragmentation # Simulations with ensemble obtained
                                                                                   # by line fragmentation
from .Simulation_dermal_filler_rheology import Simulation_dermal_filler_rheology
from .Simulation_interlocking_rheology import Simulation_interlocking_rheology

# To evaluate the stress tensors
from .EvaluationHandler import EvaluationHandler
from .EvaluationHandlerPlotter import EvaluationHandlerPlotter







from .CanvasPointsNeighbors import CanvasPointsNeighbors

class CanvasPointsShear(CanvasPointsNeighbors):
    """Canvas for placing `particleShear.CircleMassNeighbors` objects (referred to as spheres).
               This class maintains a list of objects of type
               `particleShear.CircleMassNeighbors` or derived and adds specific functionality linked to shear under
               Lees-Edwards boundary conditions"""

    def __init__(self, size_x, size_y, theCanvas=False, doDrawing=False,m=1):



        super(CanvasPointsShear,self).__init__(size_x=size_x,size_y=size_y,theCanvas=theCanvas,doDrawing=doDrawing,m=m)


        self.t=0
        """Record time while applying shear"""

        self.applyingShear = False
        """Are we currently applying a shear protocol? 
        
        Advance time only if True"""
        self.shear_rate = 0
        """Current shear rate (relative, so 1/s)"""
        self.shear = 0
        """Current applied shear (relative, so dimension-less)"""
        self.use_lees_edwards=True
        """Flag to indicate whether or not we use Lees-Edwards boundary conditions"""





    def canMove(self, theSphere):
        """Indicate whether a given sphere (object of type `particleShear.Point`
        or subclass thereof) to move on the current canvas


        This is for appyling boundary conditions where the movement is imposed externally on the top and lower edge.
        In this case all spheres can move except for in specified shear experiments where
        `particleShear.CanvasPointsShear.applyingShear` is set True\n
           This method is defined in class `particleShear.CanvasPointsShear`"""
        if not self.applyingShear:
            return True
        pos = theSphere.coordinates()
        # Spheres touching the upper and lower boundaries are attached to these boundaries and therefore immobilized
        return pos[1] > pos[2] and pos[1] < self.size_y - pos[2]


    def particle_info(self):
        return "CanvasPointsShear: Points with mass, neighbor detection and shear"



    def do_linear_acceleration(self, dt):
        """Instruct mobile spheres (objects of type `particleShear.Point` or subclasses thereof) to change their speed depending
        on the forces acting on them.

        The objects need to to derive from `particleShear.CircleMass` or
        otherwise possess a do_linear_acceleration method. In addition to its parent method
        (`particleShear.CanvasPoints.do_linear_acceleration` of class `particleShear.CanvasPoints`),
        this method imposes the local speed given by the
        current `particleShear.CanvasPointsShear.shear_rate` at the top and lower boundary to the immobile spheres



        This method is defined in class `particleShear.CanvasPointsShear`"""
        super(CanvasPointsShear,self).do_linear_acceleration(dt)
        for theSphere in self.sphereList:
            if not self.canMove(theSphere):

                pos = theSphere.coordinates()
                theSphere.yspeed = 0
                if pos[1] > self.size_y / 2:
                    theSphere.xspeed = self.shear_rate * self.size_y / 2
                else:
                    theSphere.xspeed = -self.shear_rate * self.size_y / 2

    def transmit_lees_edwards_parameters(self, theSphere):
        """Transmit the parameters pertaining to the Lees-Edwards boundary conditions to the spheres.

        If `particleShear.CanvasPointsShear.use_lees_edwards` is set to True, transmit the instance variables regarding
        canvas size (`particleShear.CanvasPoints.size_x`,`particleShear.CanvasPoints.size_y`), and shear regime
        (`particleShear.CanvasPointsShear.shear`,`particleShear.CanvasPointsShear.shear_rate`,
        `particleShear.CanvasPointsShear.use_lees_edwards`) to the given target sphere.
         Else, only set the `particleShear.PointLeesEdwards.use_lees_edwards` field in the
          target sphere to False\n
         This method is defined in class `particleShear.CanvasPointsShear`"""
        if self.use_lees_edwards:
            theSphere.size_x = self.size_x
            theSphere.size_y = self.size_y
            theSphere.shear_rate = self.shear_rate
            theSphere.shear = self.shear
            theSphere.use_lees_edwards = self.use_lees_edwards
        else:
            if hasattr(theSphere, "use_lees_edwards"):
                theSphere.use_lees_edwards = False
        return theSphere









from .Force_register import Force_register
from .Graphical_output_configuration import Graphical_output_configuration


class StressTensorEvaluation():
    """Class to evaluate mean stress tensor components from a `particleShear.Force_register`

    This class implements different expressions to evaluate mean force tensor components from the spatially localized
    force vectores as stored in  `particleShear.Force_register` variables\n
    Defined in sub-package particleShearBase"""


    def __init__(self,size_x,size_y,theCanvas=False):
        """ Initialize by setting all the mean stress tensor expression to 2x2 zero matrices"""
        self.size_x=size_x
        """ Width of the simulation area in micrometers"""
        self.size_y=size_y
        """ Height of the simulation area in micrometers"""

        self.overall_stress_tensor=[[0,0],[0,0]]
        """ Primary internal stress tensor, here implemented to be identical to 
        `particleShear.StressTensorEvaluation.stress_tensor_LW`"""


        self.stress_tensor_LW = [[0,0],[0,0]]
        """ Love-Weber definition of the stress tensor (see for instance
        eq. 18 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet, On the definition of the stress tensor 
        in granular media. International Journal of Solids and Structures, 
        2013. 50(14-15): p. 2508-2517.) Soil mechanics convention: normal is stress is positive for compression"""
        self.stress_tensor_unbalanced_forces = [[0, 0], [0, 0]]
        """ Non-compensated acceleration stress tensor (see for instance
        eq. 6 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet, On the definition of the stress tensor 
        in granular media. International Journal of Solids and Structures, 
        2013. 50(14-15): p. 2508-2517.) Soil mechanics convention: normal is stress is positive for compression"""
        self.stress_tensor_linear_acceleration_otsuki = [[0, 0], [0, 0]]
        """ Inertia compensation term given by eq. 15 of Otsuki, 
        M. and H. Hayakawa, Discontinuous change of shear modulus for frictional
        jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902.
        We do not use this term as compared to Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet, On the definition of the stress tensor 
        in granular media. International Journal of Solids and Structures, 
        2013. 50(14-15): p. 2508-2517, acceleration seems to be missing."""

        self.stress_tensor_linear_acceleration=[[0, 0], [0, 0]]
        """ Linear acceleration stress tensor. Here, we neglect the effects of gravity, so the linear 
         acceleration is linked solely to the unbalanced force (see for instance
        eq. 6 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet, On the definition of the stress tensor 
        in granular media. International Journal of Solids and Structures, 
        2013. 50(14-15): p. 2508-2517.). We consider therefore that `particleShear.StressTensorEvaluation.stress_tensor_linear_acceleration` 
         is equal `particleShear.StressTensorEvaluation.stress_tensor_unbalanced_forces`. Soil mechanics convention: 
         normal is stress is positive for compression (which would lead to acceleration away from the center of the assembly)"""


        self.stress_tensor_with_external_forces=[[0,0],[0,0]]
        """Force stress tensor ascribed to external force (see 
                eq. 6 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet, On the definition of the stress tensor 
                in granular media. International Journal of Solids and Structures, 
                2013. 50(14-15): p. 2508-2517.) Soil mechanics convention: normal is stress is positive for compression"""

        self.stress_tensor_spin_kinetic_energy=[[0,0],[0,0]]
        """Tensile stress tensor reflecting centrifugal forces from the rotation of the particles.
            This is the last of of eq. 34 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet, On the definition of the stress tensor 
            in granular media. International Journal of Solids and Structures, 
            2013. 50(14-15): p. 2508-2517.), Corrected here by 
            replacing the 2/3 coefficient by a coefficient of 1 (we think there is an error 
            in Nicot et al. regarding the inertia matrix: eq. 29 in Nicot et al. should read 
            Xij=1/5.mp.rp^2.delta_ij=1/2.Jp.delta_ij instead of Xij=2/15.mp.rp^2.delta_ij=1/3.Jp.delta_ij).\n\n 
            We also reverse the sign 
             due to using the Soil mechanics convention: normal is stress is positive for compression, so this 
            is by definition negative for rotation rates different from 0
        """

        self.stress_tensor_unbalanced_torque = [[0, 0], [0, 0]]
        """Asymmetric stress tensor reflecting the unbalanced torques causing rotational acceleration of the particles.
            This is the second last term of eq. 34 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet, 
            On the definition of the stress tensor 
            in granular media. International Journal of Solids and Structures, 
            2013. 50(14-15): p. 2508-2517.), corrected here by 
            replacing the 2/3 coefficient by a coefficient of 1 (we think there is an error 
            in Nicot et al. regarding the inertia matrix: eq. 29 in Nicot et al. should read 
            Xij=1/5.mp.rp^2.delta_ij=1/2.Jp.delta_ij instead of Xij=2/15.mp.rp^2.delta_ij=1/3.Jp.delta_ij).\n\n 
             We also reverse the sign 
             due to using the Soil mechanics convention
        """

        self.stress_tensor_internal_torque = [[0, 0], [0, 0]]
        """Asymmetric stress tensor reflecting the purely internal interaction torques 
        causing rotational acceleration of the particles.
        With restriction to the purely internal torques (vs. externally applied ones), 
        this is the second last term of eq. 34 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet, 
        On the definition of the stress tensor 
        in granular media. International Journal of Solids and Structures, 
        2013. 50(14-15): p. 2508-2517.). We corrected here the expression given by Nicot et al. by 
        replacing the 2/3 coefficient by a coefficient of 1 (we think there is an error 
        in Nicot et al. regarding the inertia matrix: eq. 29 in Nicot et al. should read 
        Xij=1/5.mp.rp^2.delta_ij=1/2.Jp.delta_ij instead of Xij=2/15.mp.rp^2.delta_ij=1/3.Jp.delta_ij) 
        We also reverse the sign 
        due to using the Soil mechanics convention
        """

        self.theCanvas=theCanvas
        """Reference to Tkinter canvas for drawing forces"""
        self.graphical_output_configuration=Graphical_output_configuration()
        """`particleShear.Graphical_output_configuration` to configure drawing"""
        self.toDeleteList=[]
        """To move lines on Tkinter displays, one needs to delete them and redraw them, this is to keep track of 
        the currently active lines"""

    def evaluate_force_stress_tensors(self, force_register):
        """Evaluate the stress tensors linked to forces

        This method evaluates the terms `particleShear.StressTensorEvaluation.stress_tensor_LW`,
        `particleShear.StressTensorEvaluation.stress_tensor_unbalanced_forces` and
        `particleShear.StressTensorEvaluation.stress_tensor_with_external_forces` from the forces and locations
        stored in the force_register argument (of type `particleShear.Force_register`)

        The `particleShear.StressTensorEvaluation.stress_tensor_LW` is evaluated from the internal
        forces acting between pairs of spheres. This information is stored in the
        `particleShear.Force_register.pair_register` field of the `particleShear.Force_register`.
        The sign convention is such that if the spheres are excerting
         each repulsive forces on each other, positive normal (diagonal) elements result in
         `particleShear.StressTensorEvaluation.stress_tensor_LW`. Since repulsion arises if the ensemble is compressed,
         this means that the sign convention associates positive normal stress with a compressive state. This is the
         Soil mechanics sign convention (Fortin, J., O. Millet, and G. de Saxce,
         Construction of an averaged stress tensor for a granular medium.
         European Journal of Mechanics a-Solids, 2003. 22(4): p. 567-582.)

         The `particleShear.StressTensorEvaluation.stress_tensor_unbalanced_forces` is evaluated from the net
         forces acting on the individual internal spheres. This information is stored in the
        `particleShear.Force_register.total_particle_force_register` field of the `particleShear.Force_register`.
         The sign convention is such that if the spheres are being pushed away from the center of the assembly,
         positive normal (diagonal) elements result in `particleShear.StressTensorEvaluation.stress_tensor_unbalanced_forces`.
         Such net repulsion arises on a compressed, but not spatially restricted assembly; a typical situation would be
          free boundaries that allow the spheres to move away from ech other and a compressed initial assembly that then
          expands into the free available space.
         Since the sign convention followed here associates positive normal stress with a compressive state, this means
          that the Soil mechanics sign convention is followed (Fortin, J., O. Millet, and G. de Saxce,
         Construction of an averaged stress tensor for a granular medium.
         European Journal of Mechanics a-Solids, 2003. 22(4): p. 567-582.

         The `particleShear.StressTensorEvaluation.stress_tensor_with_external_forces` is evaluated from the boundary
        forces acting through the domain boundaries or via spheres on the boundaries with imposed movement.
        This information is stored in the
        `particleShear.Force_register.external_force_register` field of the `particleShear.Force_register`.
        The sign convention is such that if the spheres
        are on the average pushed inward, positive normal (diagonal) elements result in
         `particleShear.StressTensorEvaluation.stress_tensor_with_external_forces`. Average inward forces
          mean that the assembly is kept in a compressive state by the external forces. Association between compressive
          state and positive normal stresses corresponds to the
         Soil mechanics sign convention (Fortin, J., O. Millet, and G. de Saxce,
         Construction of an averaged stress tensor for a granular medium.
         European Journal of Mechanics a-Solids, 2003. 22(4): p. 567-582.)

         There is a particularity when using Lees-Edwards boundary conditions and Sllod stabilization of motion
         in oscillatory shear experiments (i.e. the primary conditions used here and in
         Otsuki, M. and H. Hayakawa, Discontinuous change of shear modulus for frictional
         jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902.). The Sllod stabilization means that one
         considers only the excess momentum compared to the anticipated movement due to the general shearing of the
         assembly (see the peculiar momenta in Otsuki et al.; for the implementation of the shear speed change, see
          `particleShear.CanvasPointsBasicElasticityLeesEdwards.adjust_sphere_speed_to_shear_rate_change`
          in class  `particleShear.CanvasPointsBasicElasticityLeesEdwards`
          ). This means that the inertial forces that would be needed to accelerate
         the overall assembly to the
         local shearing speeds are ignored. For low frequencies and amplitudes, this becomes negligeable, but at higher
         shear and frequencies, the simulation will not take into account a large part of the inertial forces, whereas
         a rheological measurement would, at least before correction for inertia.

        `particleShear.StressTensorEvaluation.stress_tensor_linear_acceleration`
         is set equal to  `particleShear.StressTensorEvaluation.stress_tensor_unbalanced_forces`
         since we neglect gravity here

        Method defined in `particleShear.StressTensorEvaluation`"""
        self.stress_tensor_LW = [[0, 0], [0, 0]]
        self.stress_tensor_unbalanced_forces = [[0, 0], [0, 0]]
        self.stress_tensor_linear_acceleration = [[0, 0], [0, 0]]
        self.stress_tensor_with_external_forces = [[0, 0], [0, 0]]

        for pair in force_register.pair_register:
            target=pair[0]

            source=pair[1]

            force=pair[2]

            delta_x = [target.x-source.x, target.y-source.y]



            for i in range(2):
                for j in range(2):
                    self.stress_tensor_LW[i][j] = self.stress_tensor_LW[i][j] + force[i]*1e-12 * delta_x[j] * 1e-6 / 2

        for i in range(2):
            for j in range(2):
                self.stress_tensor_LW[i][j] = self.stress_tensor_LW[i][j]/(self.size_x*1e-6)/(self.size_y*1e-6)

        for single in force_register.total_particle_force_register:
            target = single[0]
            force = single[1]
            delta_x = [target.x-self.size_x/2 , target.y-self.size_y/2 ]



            for i in range(2):
                for j in range(2):
                    self.stress_tensor_unbalanced_forces[i][j] = self.stress_tensor_unbalanced_forces[i][j] + force[i]*1e-12 * delta_x[j] * 1e-6

        for i in range(2):
            for j in range(2):
                self.stress_tensor_unbalanced_forces[i][j] = self.stress_tensor_unbalanced_forces[i][j]/(self.size_x*1e-6)/(self.size_y*1e-6)



        if self.theCanvas:
            for theLine in self.toDeleteList:
                self.theCanvas.delete(theLine)



        for external_force in force_register.external_force_register:
            target = external_force[0]
            force = external_force[1]

            r = [target.x-self.size_x/2, target.y-self.size_y/2]

            if self.theCanvas and self.graphical_output_configuration.draw_boundary_forces:
                self.toDeleteList.append(
                    self.theCanvas.create_line(target.x,target.y,target.x+force[0]*1000,target.y+force[1]*1000,fill="darkorchid"))

            for i in range(2):
                for j in range(2):
                    self.stress_tensor_with_external_forces[i][j] = self.stress_tensor_with_external_forces[i][j]\
                                                                    -force[i]*1e-12 * r[j] * 1e-6



        for i in range(2):
            for j in range(2):
                self.stress_tensor_with_external_forces[i][j] = \
                    self.stress_tensor_with_external_forces[i][j]/(self.size_x*1e-6)/(self.size_y*1e-6)


        for i in range(2):
            for j in range(2):
                # These are synonyms as we do not include gravity
                self.stress_tensor_linear_acceleration[i][j]=self.stress_tensor_unbalanced_forces[i][j]

    def evaluate_spin_kinetic_energy_stress_tensor(self,theSphereList):
        """Evaluate the stress tensor associated with the centrifugal forces arising by rotation of the
        spheres around their axis

        The tensile stress tensor reflects the centrifugal forces from the rotation of the particles.
        This is the spin kinetic term in eq. 34 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet,
        On the definition of the stress tensor
        in granular media. International Journal of Solids and Structures,
        2013. 50(14-15): p. 2508-2517.),
        corrected by replacing the 2/3 coefficient by a coefficient of 1. See instance
        variable `particleShear.StressTensorEvaluation.stress_tensor_spin_kinetic_energy`.\n\n
        We also reverse the sign due to using the Soil mechanics convention:
        normal is stress is positive for compression, so this
        is by definition negative for rotation rates different from 0\n\n
        Method defined in `particleShear.StressTensorEvaluation`
        """



        self.stress_tensor_spin_kinetic_energy = [[0, 0], [0, 0]]

        for theSphere in theSphereList:
            # The moment of inertia is in mg*micrometers^2, but we need it in kg*m^2 here. The
            # unit conversion factor is therefore 10^(-18)

            moment_of_inertia = theSphere.inertia * 1e-18
            K = theSphere.omega *theSphere.omega / 2 * moment_of_inertia
            for i in range(2):
                self.stress_tensor_spin_kinetic_energy[i][i] = self.stress_tensor_spin_kinetic_energy[i][i] - K
        # Now, with respect to the sampling volume
        for i in range(2):
            for j in range(2):
                self.stress_tensor_spin_kinetic_energy[i][j] = self.stress_tensor_spin_kinetic_energy[i][j] \
                                                               / (self.size_x * 1e-6) / (self.size_y * 1e-6)


    def evaluate_unbalanced_torque_stress_tensor(self,force_register):
        """Evaluate the stress tensor associated with overall unbalanced torques acting on the particle

         The unbalanced torque stress tensor reflects the asymmetric contribution of torques leading
          to acceleration or deceleration of the rotation of the particles.
          This is the moment term in eq. 34 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet,
          On the definition of the stress tensor
          in granular media. International Journal of Solids and Structures,
          2013. 50(14-15): p. 2508-2517.),
          corrected by replacing the 2/3 coefficient by a coefficient of 1. See instance
          variable `particleShear.StressTensorEvaluation.stress_tensor_unbalanced_torque`.\n\n
          We also reverse the sign due to using the Soil mechanics convention, implying sign reversal as
          compared to Nicot et al.\n\n
          Method defined in `particleShear.StressTensorEvaluation`
          """


        self.stress_tensor_unbalanced_torque= [[0, 0], [0, 0]]

        # The spring constant is in mg/s-2/m of depth, and is multiplied with micrometers to get the forces
        # So the force have the units micrometers*mg/s-2/m
        # For the torque, this is further multiplied by micrometers, so that makes
        # micrometers^2/mg/s-2/m of depth

        for unbalanced_torque in force_register.unbalanced_moment_register:
            M = unbalanced_torque[1]*1e-18
            self.stress_tensor_unbalanced_torque[0][1]=self.stress_tensor_unbalanced_torque[0][1]-M/2
            self.stress_tensor_unbalanced_torque[1][0] =self.stress_tensor_unbalanced_torque[1][0]+M/2
            # Now, with respect to the sampling volume
        for i in range(2):
            for j in range(2):
                self.stress_tensor_unbalanced_torque[i][j] = self.stress_tensor_unbalanced_torque[i][j] \
                                                               / (self.size_x * 1e-6) / (self.size_y * 1e-6)

    def evaluate_internal_torque_stress_tensor(self,force_register):
        """Evaluate the stress tensor associated with internal unbalanced torques acting on the particle

        The internal unbalanced torque stress tensor reflects the asymmetric contribution of torques arising
        from frictional particle-particle interaction strictly within the simulation area.
        This is the moment term in eq. 34 of Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet,
        On the definition of the stress tensor
        in granular media. International Journal of Solids and Structures,
        2013. 50(14-15): p. 2508-2517.), restricted to the internal torques, and
        corrected by replacing the 2/3 coefficient by a coefficient of 1. See instance
        variable `particleShear.StressTensorEvaluation.stress_tensor_internal_torque`.\n\n
        We also reverse the sign due to using the Soil mechanics convention, implying sign reversal as
        compared to Nicot et al.\n\n
        Method defined in `particleShear.StressTensorEvaluation`
        """


        self.stress_tensor_internal_torque= [[0, 0], [0, 0]]

        # The spring constant is in mg/s-2/m of depth, and is multiplied with micrometers to get the forces
        # So the force have the units micrometers*mg/s-2/m
        # For the torque, this is further multiplied by micrometers, so that makes
        # micrometers^2/mg/s-2/m of depth

        for internal_torque in force_register.internal_moment_register:
            M = internal_torque[2]*1e-18
            self.stress_tensor_internal_torque[0][1]=self.stress_tensor_internal_torque[0][1]-M/2
            self.stress_tensor_internal_torque[1][0] =self.stress_tensor_internal_torque[1][0]+M/2
            # Now, with respect to the sampling volume
        for i in range(2):
            for j in range(2):
                self.stress_tensor_internal_torque[i][j] = self.stress_tensor_internal_torque[i][j] \
                                                               / (self.size_x * 1e-6) / (self.size_y * 1e-6)

    def evaluate_rotational_stress_tensors(self,force_register,theSphereList):
        """Evaluate the rotational stress tensors

        This function invokes `particleShear.StressTensorEvaluation.evaluate_spin_kinetic_energy_stress_tensor`,
        `particleShear.StressTensorEvaluation.evaluate_unbalanced_torque_stress_tensor` and
        `particleShear.StressTensorEvaluation.evaluate_internal_torque_stress_tensor`. The results are stored in the
        instance variables `particleShear.StressTensorEvaluation.stress_tensor_spin_kinetic_energy`, `particleShear.StressTensorEvaluation.stress_tensor_unbalanced_torque`
        and `particleShear.StressTensorEvaluation.stress_tensor_internal_torque` \n\n
        Method defined in `particleShear.StressTensorEvaluation`
        """

        self.evaluate_spin_kinetic_energy_stress_tensor(theSphereList)
        self.evaluate_unbalanced_torque_stress_tensor(force_register)
        self.evaluate_internal_torque_stress_tensor(force_register)



    def evaluate_peculiar_momentum_transport_for_linear_acceleration_tensor(self,theSphereList,shear_rate):
        """ Evaluate the inertial compensation term defined by Otsuki et al.

        This term is the inertial compensation term (second righthand term) in eq. 15 of Otsuki, M. and H. Hayakawa,
        Discontinuous change of shear modulus for frictional
        jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902. We provide this term for completeness. This term
        does not appear as such in other main references on shear tensors in granular media (Nicot, F., N. Hadda,
        M. Guessasma, J. Fortin, and O. Millet,
        On the definition of the stress tensor in granular media.
        International Journal of Solids and Structures, 2013. 50(14-15): p. 2508-2517.), and it seems to have
        unphysical aspects. Consider a simplified assembly made from a single sphere without applied shear,
        moving linearly without applied force. All elements of the stress tensor should be zero for such a freely moving
        sphere free from any applied external force.
        If the speed of the sphere is not exactly along the x- or y-direction, the inertial compensation term by Otsuki et al.
        is non-zero, and this should not be.
        So while we provide the inertial term by Otsuki et al. for completeness, we do not
        recommend using it.

        Method defined in `particleShear.StressTensorEvaluation`"""

        self.stress_tensor_linear_acceleration_otsuki = [[0, 0], [0, 0]]


        for theSphere in theSphereList:
            # Units: forces are in N
            # the mass is in mg
            # the length scales are in micrometer
            # so here to kgm/s one needs to multiply by 10^-12
            px=(theSphere.xspeed-shear_rate*(theSphere.y-self.size_y/2))*theSphere.m*1e-12
            py=theSphere.yspeed*theSphere.m*1e-12
            p=[px,py]

            for i in range(2):
                for j in range(2):
                    # Eq. 15 of Otsuki, M. and H. Hayakawa, Discontinuous change of shear modulus for frictional
                    # jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902.
                    # This doesn't seem to be quite right as it doesn't seem to involve acceleration
                    self.stress_tensor_linear_acceleration_otsuki[i][j] = self.stress_tensor_linear_acceleration_otsuki[i][j]-\
                                                                   p[i]*p[j]/theSphere.m*1e6


        for i in range(2):
            for j in range(2):
                self.stress_tensor_linear_acceleration_otsuki[i][j] = \
                    self.stress_tensor_linear_acceleration_otsuki[i][j]/(self.size_x*1e-6)/(self.size_y*1e-6)


    def evaluate_stress_tensors(self,force_register,theSphereList,shear_rate):
        """Evaluate the stress tensor expressions

        This is the main method to be called for stress tensor evaluation. It uses the forces stored in the various
        fields of the force_register to evaluate the stress tensor expressions linked to forces via
        `particleShear.StressTensorEvaluation.evaluate_force_stress_tensors`. The method also calls
        `particleShear.StressTensorEvaluation.evaluate_peculiar_momentum_transport_for_linear_acceleration_tensor`,
        see these two methods for further details.

        Finally, the relevant overall internal force stress tensor
        (`particleShear.StressTensorEvaluation.overall_stress_tensor`) is set equal to the Love-Weber expression. This
         intrinsically compensates for inertia by local acceleration, but it
        ignores the spin terms (rotation of the spheres and change in rotation rate). The approximation is justified by
        the quasistatic conditions used in the simulations, but may need improved implementation in the future for
        simulations with rapidly spinning ("rolling") particles.

        Method defined in `particleShear.StressTensorEvaluation`"""


        self.evaluate_force_stress_tensors(force_register)
        self.evaluate_peculiar_momentum_transport_for_linear_acceleration_tensor(theSphereList,shear_rate)
        self.evaluate_rotational_stress_tensors(force_register,theSphereList)

        self.overall_stress_tensor=[[0,0],[0,0]]
        for i in range(2):
            for j in range(2):
                # Compensation for linear acceleration as in Nicot, F., N. Hadda, M. Guessasma, J. Fortin, and O. Millet, On
                # the definition of the stress tensor in granular media.
                # International Journal of Solids and Structures, 2013. 50(14-15): p. 2508-2517.
                # Also, taking into account only internal torque
                self.overall_stress_tensor[i][j] = self.overall_stress_tensor[i][j]+self.stress_tensor_LW[i][j] \
                                                   +self.stress_tensor_internal_torque[i][j]\
                                                   +self.stress_tensor_spin_kinetic_energy[i][j]




    def evaluate_quasistatic_shear_stress(self):
        """Evaluate the internal shear stress

        From the stress tensor (`particleShear.StressTensorEvaluation.overall_stress_tensor`) calculated from the
        internal forces, evaluate the shear stress (xy element). Since an external measurement apparatus would
        have to provide a force opposing the shear stress force arising at the surface, we invert signs and return the
        negative of the xy element of `particleShear.StressTensorEvaluation.overall_stress_tensor`.

        Method defined in `particleShear.StressTensorEvaluation`"""


        return (-self.overall_stress_tensor[0][1])

    def evaluate_externally_applied_shear_stress(self):
        """Evaluate the shear stress from the external forces acting on the ensemble

        From the external force stress tensor
        (`particleShear.StressTensorEvaluation.stress_tensor_with_external_forces`) calculated from the
        externally acting forces, evaluate the shear stress (xy element). To account for the fact that
        the force on an external measurement apparatus would be the inverse of the force applied on the
        boundary paraticles, we return the negative of the xy element of the
        `particleShear.StressTensorEvaluation.stress_tensor_with_external_forces`.

         Method defined in `particleShear.StressTensorEvaluation`"""


        return (-self.stress_tensor_with_external_forces[0][1])















"""Provide 2D simulations of frictional and non-frictional, crosslinked and non-crosslinked
 spheres under oscillatory shear.

The package provides the tools necessary to simulate ensembles of discrete spherical particles as described in
Otsuki, M. and H. Hayakawa, Discontinuous change of
shear modulus for frictional jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902. In addition, the particles
can be crosslinked permanently between neighbors to constitute particles\n
Package written by Thomas Braschler (thomas.braschler@gmail.com)
"""

__all__ = ["Force_register","StressTensorEvaluation","Graphical_output_configuration","neighbor_relation",
           "Point","PointLeesEdwards",
           "Circle","CircleBasicElasticity",
           "CircleFrictionElasticity","CircleMass","CircleMassNeighbors",
           "CanvasPoints","CanvasPointsBasicElasticity","CanvasPointsBasicElasticityLeesEdwards",
           "CanvasPointsFrictionElasticityLeesEdwards","CanvasPointsMass","CanvasPointsShear","CanvasPointsNeighbors",
           "elastic_force_law_plateau","elastic_force_law","PlateauConfiguration","distance_transform_plateau",
           "Sphere","SphereLeesEdwards","SphereFriction","SphereFrictionLeesEdwards",
           "Ensemble",
           "EnsembleFriction",
           "r_estimate","EnsembleLeesEdwards","EnsembleFrictionLeesEdwards",
           "particle_shear_model_parameters",
           "neighbor_relation_linkable",
           "CanvasPointsLinkable", "SphereLinkable", "SphereLinkableAdjustableInterfaceStrength",
           "EnsembleLinkable", "EnsembleCompactParticles",
           "EnsembleCompactParticlesAdjustableInterfaceStrength",
           "EnsembleCompactParticlesAdjustableInterfaceStrengthFromModelParameters",
           "EnsembleCompactParticlesFromModelParameters",
           "SphereLinkable", "SphereLinkableAdjustableInterfaceStrength",
           "elastic_force_law_tensile","elastic_force_law_tensile_exponential",
           "TensileConfiguration",
           "OscillatoryShearExperiment",
           "OscillatorySimulation",
           "OscillatorySimulationFragmentation","Simulation_dermal_filler_rheology",
           "Simulation_interlocking_rheology",
           "EvaluationHandler",
           "EvaluationHandlerPlotter",
           "doParticleShearSimulation",
           "doParticleShearSimulationSeries"
           ]



# Basic tools
from particleShearBase import *
from particleShearLinkableObjects import *
from particleShearObjects import *
from particleShearSimulation import *
from particleShearRunSimulation import *




import unittest
from particleShear import *
from tkinter import *
import math


def getSimulationCrosslinkedShear(theAmplitude=0.2,theMu=0.01,
                    cut_lines=5,Young_modulus_spheres=8000,
                    N=150,packing_fraction=1.5,density=50,bimodal_factor=1.4,
                    relative_viscosity=0.1,relative_transversal_link_strength=1,relative_frequency=0.025,
                    avoid_horizontal_angle_degree=15,interface_reenforcement_central=1,interface_reenforcement_tangential=1,
                    keep_viscosity_coefficients_constant=False,avoid_height_spanning_particles=True,cut_top_bottom=False,
                    imageFileType ="jpg",
                    remove_link_fraction=0,edge_fuzziness=0,
                    central_repulsion_coefficient=1,
                    doCutByTriangulation=True,doDrawing=True,relative_y_scale_force = 1e8,
                    pre_periods=2,periods=3,post_periods=1,cool_factor=0.5):



    root_folder = ""
    do_permanent_links = True
    saveOutputImages = False
    saveData = False



    theSimulation=Simulation_interlocking_rheology(
        root_folder, do_permanent_links=do_permanent_links, cut_lines=cut_lines, N=N,
        packing_fraction=packing_fraction,
        density=density,
        bimodal_factor=bimodal_factor,
        amplitude=theAmplitude, Young_modulus_spheres=Young_modulus_spheres, mu=theMu,
        relative_y_scale_force=relative_y_scale_force, relative_frequency=relative_frequency, relative_viscosity=relative_viscosity,
        relative_transversal_link_strength=relative_transversal_link_strength,
        central_repulsion_coefficient=central_repulsion_coefficient,
        avoid_horizontal_angle_degree=avoid_horizontal_angle_degree,
        interface_reenforcement_central=interface_reenforcement_central,
        interface_reenforcement_tangential=interface_reenforcement_tangential,
        keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant,
        avoid_height_spanning_particles=avoid_height_spanning_particles,
        cut_top_bottom=cut_top_bottom,
        doCutByTriangulation=doCutByTriangulation,
        remove_link_fraction=remove_link_fraction,
        edge_fuzziness=edge_fuzziness,
        doDrawing=doDrawing,
        saveData=saveData
    )
    theSimulation.function_call = theSimulation.function_call + \
                                  "\n\ttheSimulation.baseline_pre_periods = " + str(pre_periods) + \
                                  "\n\ttheSimulation.baseline_post_periods = " + str(post_periods) + \
                                  "\n\ttheSimulation.saveOutputImages = " + str(saveOutputImages) + \
                                  "\n\ttheSimulation.imageFileType = " + str(imageFileType) + \
                                  "\n\ttheSimulation.runSimulation(periods=" + str(periods) + ", cool_factor=" + \
                                  str(cool_factor) + ")"

    theSimulation.baseline_pre_periods = pre_periods
    theSimulation.baseline_post_periods = post_periods

    theSimulation.saveOutputImages = saveOutputImages
    theSimulation.imageFileType = imageFileType

    return(theSimulation)


def initiateSimulationCrosslinkedShear(theSimulation, cool_factor=1, periods=2, plotStress=True):

    theSimulation.initEnsemble(do_debug=True)

    model = theSimulation.theEnsemble.model

    y_scale_force = theSimulation.relative_y_scale_force / model.k / theSimulation.amplitude

    final_dt = theSimulation.theEnsemble.dt

    # if theSimulation.amplitude<0.01:

    #    final_dt=final_dt/(1-2*(math.log10(theSimulation.amplitude)+2))
    #    print("Small amplitude: Adjusted to dt=", final_dt)

    print("Setting relative frequency: " + str(theSimulation.relative_frequency))
    theSimulation.f = theSimulation.relative_frequency / model.time_constant

    fileName = theSimulation.file_path()

    if theSimulation.saveOutputImages:
        theSimulation.imageOutputFolder = theSimulation.image_folder_path(fileName)
        theSimulation.imageBaseFileName = "image"

    theFile = False
    if theSimulation.saveData:
        theFile = open(fileName, "w")

    theSimulation.write_information_on_simulation_to_file(theFile)

    print("Creating oscillatory shear experiment, dt = ", final_dt, " max dt = ", theSimulation.theEnsemble.dt_max)
    theExperiment = OscillatoryShearExperiment(theSimulation.theEnsemble, final_dt,
                                               theSimulation.f, theSimulation.amplitude, theFile,
                                               force_scale=y_scale_force,
                                               TkSimulation=theSimulation.theEnsemble.theTkSimulation,
                                               dt_max=theSimulation.theEnsemble.dt_max,
                                               imageOutputFolder=theSimulation.imageOutputFolder,
                                               imageBaseFileName=theSimulation.imageBaseFileName,
                                               imageFileType=theSimulation.imageFileType,
                                               plotStress=plotStress)

    theExperiment.periods = periods
    theExperiment.baseline_pre_periods = theSimulation.baseline_pre_periods
    theExperiment.baseline_post_periods = theSimulation.baseline_post_periods

    theExperiment.do_preequilibration_with_fixed_boundaries = True

    N_pre_equilibration_fixed_boundaries = 25 * theSimulation.adjustment_factor_for_bounded_preequilibration_steps()

    N_per_period = 1 / theSimulation.f / final_dt

    cf_per_dt = math.pow(cool_factor, 1 / N_per_period)

    # The idea here is that we have always the same attenuation per period
    print("cool factor per dt=", cf_per_dt)

    theSimulation.theExperiment = theExperiment

    theSimulation.cool_factor_per_dt = cf_per_dt




class TestSimulationCrosslinkedShear(unittest.TestCase):

    def setUp(self):

        self.imageOutputFolder=False

        self.theSimulation = getSimulationCrosslinkedShear(cut_lines=3,N=60,pre_periods=0.1)

        initiateSimulationCrosslinkedShear(self.theSimulation)


        self.theEnsemble = self.theSimulation.theEnsemble

        self.theExperiment = self.theSimulation.theExperiment


        N_pre_equilibration_fixed_boundaries=25

        self.theExperiment.final_cool_factor = 1

        self.theEvaluator = StressTensorEvaluation(self.theEnsemble.size_x, self.theEnsemble.size_y,
                                                   self.theEnsemble.theCanvas)


        self.theExperiment.pre_equilibrate(N=N_pre_equilibration_fixed_boundaries)

        self.theEnsemble.t = 0

        self.theEnsemble.applyingShear = True



        if self.theExperiment.plotStress and not self.theExperiment.theTk:
            self.theExperiment.theTk = Tk()

        if self.theExperiment.plotStress:
            self.theExperiment.outputCanvas = Canvas(
                self.theExperiment.theTk, width=self.theExperiment.width, height=self.theExperiment.height)

            self.theExperiment.outputCanvas.pack()

        self.theEnsemble.t = 0


        Time_required = (self.theExperiment.periods + self.theExperiment.baseline_pre_periods +\
                         self.theExperiment.baseline_post_periods) / self.theExperiment.f

        x_scale = self.theExperiment.width / Time_required

        self.theExperiment.plotter = EvaluationHandlerPlotter(self.theExperiment.theTk, self.theExperiment.outputCanvas, x_scale=x_scale,
                                                y_scale_force=self.theExperiment.force_scale,
                                                y_scale_strain=30 / self.theExperiment.amplitude, r=2, offset_x=0,
                                                offset_y=self.theExperiment.height / 2,
                                                plotStress=self.theExperiment.plotStress)

        N = int((self.theExperiment.periods*0.15 + self.theExperiment.baseline_pre_periods
                         ) / self.theExperiment.f / self.theExperiment.dt)

        print("Start application shear protocol, total N=", N, " steps")






        self.theEnsemble.t = 0



        for i in range(N):
            if i >= (self.theExperiment.baseline_pre_periods / self.theExperiment.f / self.theExperiment.dt) and i <= (
                    (self.theExperiment.periods + self.theExperiment.baseline_pre_periods) / self.theExperiment.f / self.theExperiment.dt):
                self.theEnsemble.setShearRate(
                    self.theExperiment.amplitude * self.theExperiment.out_of_phase_signal(self.theEnsemble.t) *
                    self.theExperiment.f * 2 * math.pi)
            else:
                self.theEnsemble.setShearRate(0)

            # Calculate forces, this registers individual forces in the force register

            self.theEnsemble.mechanical_simulation_step_calculate_forces()

            # report net particles forces and moments in the force register
            self.theEnsemble.record_total_particle_forces()

            self.theExperiment.plotter.record(self.theEnsemble, self.theEvaluator)

            # Do linear and rotational acceleration
            self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=
                                                                               self.theSimulation.cool_factor_per_dt,
                                                                               dt=self.theExperiment.dt)

            # Do linear and rotation movememnt
            self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.theExperiment.dt)
            # reset forces for next round
            self.theEnsemble.reset_force()

            if self.theEnsemble.doDrawing:
                self.theExperiment.theTkSimulation.update()

        self.plotter = self.theExperiment.plotter








    def tearDown(self):
        if self.theSimulation.theEnsemble.theTkSimulation:
            self.theSimulation.theEnsemble.theTkSimulation.destroy()

        if self.theExperiment.theTk:
            self.theExperiment.theTk.destroy()

        self.theSimulation.theTkSimulation = False
        self.theExperiment.theTk = False



    def test_stress_tensors(self):


        theEvaluator=self.theEvaluator


        self.assertTrue(theEvaluator.stress_tensor_LW[0][0]>500 and theEvaluator.stress_tensor_LW[1][1]>500)

        self.assertTrue(abs(theEvaluator.overall_stress_tensor[0][1]-theEvaluator.overall_stress_tensor[1][0])/
              (abs(theEvaluator.overall_stress_tensor[0][0])+abs(theEvaluator.overall_stress_tensor[1][1]))<1e-15)


        linear_acceleration=self.theEvaluator.stress_tensor_linear_acceleration
        self.assertTrue(abs(linear_acceleration[0][0])<0.1*abs(theEvaluator.overall_stress_tensor[0][0]) and
                        abs(linear_acceleration[0][1]) < 0.1 * abs(theEvaluator.overall_stress_tensor[0][0]) and
                        abs(linear_acceleration[1][0]) < 0.1 * abs(theEvaluator.overall_stress_tensor[0][0]) and
                        abs(linear_acceleration[1][1]) < 0.1 * abs(theEvaluator.overall_stress_tensor[0][0]))

        external = self.theEvaluator.stress_tensor_with_external_forces
        self.assertTrue(abs((external[0][0]-theEvaluator.overall_stress_tensor[0][0])/theEvaluator.overall_stress_tensor[0][0])<0.1 and
                        abs((external[1][1]-theEvaluator.overall_stress_tensor[1][1])/theEvaluator.overall_stress_tensor[1][1]) < 0.1)

        print("external force")
        print(abs((external[0][0]-theEvaluator.overall_stress_tensor[0][0])/theEvaluator.overall_stress_tensor[0][0]))


if __name__ == '__main__':
    unittest.main()


import unittest
from particleShear import *
from tkinter import *
import math



class TestoneSphereFreeOneSphereBoundary(unittest.TestCase):

    def setUp(self):



        size_x = 500
        size_y = 500

        theTk = Tk()
        theCanvas = Canvas(theTk, width=size_x, height=size_y)

        theCanvas.pack()

        theEnsemble = EnsembleLinkable(500, 500, 0, 0.8, theCanvas, True, k=1, nu=1, k_t=1, nu_t=1, mu=0.5)

        theEnsemble.sphereList.append(SphereLinkable("GREY", 150, 80, 120, m=1, my_index=1, theCanvas=theCanvas,
                                                     doDrawing=True, force_register=theEnsemble,
                                                     size_x=theEnsemble.size_x,
                                                     size_y=theEnsemble.size_y))

        theEnsemble.sphereList[0].xspeed = 10

        theEnsemble.sphereList.append(SphereLinkable("BLUE", 300, 10, 120, m=1, my_index=2, theCanvas=theCanvas,
                                                     doDrawing=True, force_register=theEnsemble,
                                                     size_x=theEnsemble.size_x,
                                                     size_y=theEnsemble.size_y))

        theEnsemble.sphereList[1].xspeed = -10

        theEnsemble.sphereList[1].omega = 0

        theEnsemble.applyingShear=True

        graphical = theEnsemble.graphical_output_configuration

        graphical.draw_rotation_line = True

        graphical.color_spheres_boundary="Green"

        graphical.draw_permanent_connections=True

        theEnsemble.set_graphical_output_configuration(graphical)

        #theEnsemble.sphereList[0].establish_permanent_link(theEnsemble.sphereList[1])

        theTk.update()

        theEvaluator = StressTensorEvaluation(size_x, size_y)

        self.theEvaluator = theEvaluator
        self.theEnsemble = theEnsemble
        self.theTk=theTk




        self.dt=0.1
        self.cool_factor=1

        for i in range(70):
            self.theEnsemble.mechanical_simulation_step_calculate_forces()

            self.theEnsemble.record_total_particle_forces()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)


            self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
            self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
            self.theTk.update()


    def tearDown(self):
        self.theTk.destroy()
        self.theTk = False





    def test_stress_tensor_LW(self):






        LW = self.theEvaluator.stress_tensor_LW





        self.assertTrue(LW[0][0]==0 and LW[0][1]==0 and LW[1][0]==0 and LW[1][1]==0)


    def test_overall_stress_tensor(self):



        overall = self.theEvaluator.overall_stress_tensor

        # Overall should be diagonal, with equal, negative diagonal elements

        self.assertTrue(overall[0][0]==overall[1][1])

        self.assertTrue(overall[0][1] == 0 and overall[1][0]==0)



    def test_external_stress_tensor(self):


        external = self.theEvaluator.stress_tensor_with_external_forces


        self.assertTrue(
            abs(external[0][0]) > 0 or abs(external[0][1]) > 0 or abs(external[1][0])>0
            or abs(external[1][1])>0)

    def test_unbalanced_forces_stress_tensor(self):




        unbalanced = self.theEvaluator.stress_tensor_unbalanced_forces

        external = self.theEvaluator.stress_tensor_with_external_forces


        external_tot = abs(external[0][0])+abs(external[0][1])+abs(external[1][0])+abs(external[1][1])


        self.assertTrue(
            abs(external[0][0] +unbalanced[0][0])/external_tot <1e-15)








if __name__ == '__main__':
    unittest.main()


from .Ensemble import r_estimate
import math

class particle_shear_model_parameters():
    """Model class to facilitate calculation of particle and simulation parameters from physical properties of the
        constituent spheres

        Class defined in subpackage particleShearObjects"""
    def __init__(self,size_x=500,size_y=500,N=100,packing_fraction=1,Young_modulus_spheres=10000,density=50,
                 bimodal_factor=1):
        """Initialize model

        Parameters\n
        `size_x` Width of area to be used in pixels = micrometers for the simulation\n
        `size_y` Height of area to be used in pixels = micrometers for the simulation\n
        `N` The number of spheres to place\n
        `packing_fraction` The packing fraction indicating the density; this is defined as the area occupied by the
         non-compressed spheres as compared to the actual available area (size_x times size_y)\n
         `Young_modulus_spheres` The Young modulus of the constituent spheres in Pa\n
         `density of the spheres in kg/m^3\n`
         `bimodal_factor` is ratio of radii of the smaller and larger spheres in the bimodal distribution
          """

        self.bimodal_factor=bimodal_factor
        """Bimodal factor indicating the ratio between the larger and the smaller spheres (bimodal size distribution)"""
        self.size_x=size_x
        """Width of the simulation area in micrometers (=Pixels) """
        self.size_y=size_y
        """Height of the simulation area in micrometers (=Pixels) """
        self.size_x_m=size_x/1e6 # Pixel units are micrometers
        """Width of the simulation area in meters """
        self.size_y_m=size_y/1e6 # Pixel units are micrometers
        """Height of the simulation area in meters """
        self.N=N
        """Number of spheres"""
        self.packing_fraction=packing_fraction
        """Packing fraction: relative area occupied by the uncompressed spheres as compared to the available area"""
        self.r=r_estimate(self.size_x,self.size_y,self.packing_fraction,self.N,
                          bimodal_upper=math.sqrt(bimodal_factor),bimodal_lower=1/math.sqrt(bimodal_factor))
        """Average radius in micrometers.
        
        The radius is defined such that the larger radius would be sqrt(bimodal_factor)*radius and the smaller
        1/sqrt(bimodal_factor)*radius. The expected packing fraction is conserved by adjusting for the increase in 
        average cross section area resulting from the use of a bimodal distribution"""
        self.r_m=self.r/1e6 # Pixel units are micrometers
        """Average radius in meters"""
        self.m=math.pi*self.r_m*self.r_m*density*1e6 #This is per m of depth for the cylinders; the mass is converted to mg/m
        """Average mass per particle, in mg/m (cylindrical particles shown in cross section, mass per m of cylinder depth)"""

        self.Young_modulus_spheres = Young_modulus_spheres
        """Young modulus of the constituent spheres; this is of the material of which the spheres are made, in general, 
        the ensemble would be softer"""
        self.density=density
        """Density of the particles, in kg/m^3"""

        self.k=math.pi/8*Young_modulus_spheres*1e6
        """Spring constant, in mg/s^2/m; this relates the force per unit length to the indentation depth.
        
        The formula used is pi/8*Young_modulus spheres; this is eq. 5.34 from Popov, V.L, 
        Contact Mechanics and Friction: Physical Principles and Applications, Berlin, Germany, 
        2017, 2nd edition (DOI 10.1007/978-3-662-53081-8), with an effective modulus given by 
         E*=E/2 from eq. 5.26 also in Popov's contact mechanics book. We neglect here the effect of the 
          Poisson coefficient (i.e. we assume Poisson coefficient=0), this introduces at worst 25% error"""


        self.time_constant=math.sqrt(self.m/self.k)/math.sqrt(max(bimodal_factor,1/bimodal_factor))
        """Characteristic time constant sqrt(m/k)
        
        
        Here, the expressions of k and m in per m of depth cancels out
        Also, regarding a possible bimodal distribution (bimodal_factor>1), we err on the side of caution
        If a bimodal distribution is provided, then, the smaller radius will be r/sqrt(bimodal_factor), and so the mass
        will be m/bimodal_factor. That makes the time constant smaller, by sqrt(bimodal_factor). To be sure against
        indication of bimodal factor < 1, use the maximum of (bimodal_factor, 1/bimodal_factor)
        """



from particleShearObjects import particle_shear_model_parameters
from .OscillatoryShearExperiment import OscillatoryShearExperiment
import math
from tkinter import *
from pathlib import Path
import os



class OscillatorySimulation():

    def __init__(self,root_folder,amplitude=0.5,relative_y_scale_force=5e-4,relative_frequency=0.05,doDrawing=True,
                 saveData=False):
        self.amplitude=amplitude
        self.doDrawing=doDrawing
        self.relative_y_scale_force=relative_y_scale_force
        self.relative_frequency=relative_frequency
        self.theTkSimulation=False
        self.f=-1 # Frequency, -1 before setting
        self.root_folder=root_folder
        self.saveData=saveData

        self.baseline_pre_periods = 0  # Acquisition of stress data with 0
        # excitation amplitude to have a baseline before oscillatory shear
        self.baseline_post_periods = 0  # Acquisition of stress data with 0
        # excitation amplitude to have a baseline before oscillatory shear

        self.imageOutputFolder = False

        self.imageBaseFileName = False

        self.saveOutputImages = False

        self.imageFileType="jpg"

        self.saveStressTensorData=True
        """Flag indicating whether we should save the detailed stress tensor data in the ASCII output files
        
        Relevant only if saveData==True"""

        self.function_call = "OscillatorySimulation(root_folder="+str(root_folder)+",amplitude="+str(amplitude)+\
            ",relative_y_scale_force="+str(relative_y_scale_force)+",relative_frequency="+str(relative_frequency)+")"

        self.theExperiment = False # To hold the shear experiment done

    # Here the ensemble needs to be initiated, but this needs to be done in the subclasses
    def initEnsemble(self):
        if not self.theTkSimulation and self.doDrawing:
            self.theTkSimulation=Tk()


    def file_path(self):
        if not self.root_folder:
            return False
        fname = "oscillatory_simulation_"

        fnum=1

        while Path(self.root_folder + "/" + fname + str(fnum) + ".txt").is_file():
            fnum = fnum + 1

        return self.root_folder + "/" + fname + str(fnum) + ".txt"

    def image_folder_path(self,fileName):
        folderName=fileName+"_images_folder"
        self.createFolder(folderName)
        return folderName

    def createFolder(self,directory):
        print("Creating image folder",directory)
        if not os.path.exists(directory):
            os.makedirs(directory)
            print("done")








    def write_information_on_simulation_to_file(self,theFile):
        if not self.saveData:
            return
        theFile.write("Basic oscillatory shear simulation")

    # To be implemented by subclasses, in some cases longer pre-equilibration is necessary (low friction coefficient, low amplitudes)
    # but this depends on the exact simulation to be run
    def adjustment_factor_for_bounded_preequilibration_steps(self):
        return 1


    def runSimulation(self,cool_factor=1,periods=2,plotStress=True):



        self.initEnsemble()




        model = self.theEnsemble.model

        y_scale_force = self.relative_y_scale_force / model.k / self.amplitude

        final_dt = self.theEnsemble.dt

        #if self.amplitude<0.01:

        #    final_dt=final_dt/(1-2*(math.log10(self.amplitude)+2))
        #    print("Small amplitude: Adjusted to dt=", final_dt)




        print("Setting relative frequency: "+str(self.relative_frequency))
        self.f = self.relative_frequency / model.time_constant

        fileName=self.file_path()

        if self.saveOutputImages:
            self.imageOutputFolder = self.image_folder_path(fileName)
            self.imageBaseFileName = "image"

        theFile=False
        if self.saveData:
            theFile = open(fileName, "w")



        self.write_information_on_simulation_to_file(theFile)

        print("Creating oscillatory shear experiment, dt = ",final_dt," max dt = ",self.theEnsemble.dt_max)
        theExperiment=OscillatoryShearExperiment(self.theEnsemble,final_dt,
                                         self.f,self.amplitude,theFile,force_scale=y_scale_force,
                                         TkSimulation=self.theEnsemble.theTkSimulation,dt_max=self.theEnsemble.dt_max,
                                         imageOutputFolder=self.imageOutputFolder,
                                         imageBaseFileName=self.imageBaseFileName,
                                         imageFileType=self.imageFileType,
                                         plotStress=plotStress,
                                         saveStressTensorData=self.saveStressTensorData)

        theExperiment.periods=periods
        theExperiment.baseline_pre_periods=self.baseline_pre_periods
        theExperiment.baseline_post_periods=self.baseline_post_periods






        theExperiment.do_preequilibration_with_fixed_boundaries=True

        N_pre_equilibration_fixed_boundaries=25*self.adjustment_factor_for_bounded_preequilibration_steps()


        N_per_period = 1/self.f/final_dt

        cf_per_dt = math.pow(cool_factor,1/N_per_period)

        # The idea here is that we have always the same attenuation per period
        print("cool factor per dt=",cf_per_dt)

        G=theExperiment.oscillatory_shear_experiment(
            cool_factor=cf_per_dt,N_pre_equilibration_fixed_boundaries=N_pre_equilibration_fixed_boundaries)


        if self.saveData:
            theFile.close()

        if  self.theEnsemble.theTkSimulation:
            self.theEnsemble.theTkSimulation.destroy()

        self.theTkSimulation=False

        self.theExperiment=theExperiment

        return(G)









from tkinter import *
import random
import math
from particleShearBase import CircleFrictionElasticity
from particleShearBase import Force_register

class SphereFrictionLeesEdwards(CircleFrictionElasticity):
     """Objects of type `particleShear.CircleFrictionElasticity` , using the periodic Lees-Edwards boundary conditions"""
     def __init__(self, color, x, y, diameter, m=1,my_index=1, theCanvas=FALSE, doDrawing=FALSE,
                 force_register=Force_register(),
                 size_x=500,size_y=500):
        """Constructor as `particleShear.CircleFrictionElasticity.__init__` in `particleShear.CircleFrictionElasticity`
        but with and indexing element (my_index) and with the parameter use_lees_edwards set to True"""
        super(SphereFrictionLeesEdwards,self).__init__(color, x, y, diameter, m, theCanvas, doDrawing,force_register,
                                                       use_lees_edwards=True)


        self.use_lees_edwards=True
        self.myindex = my_index  # To give a unique index
        self.size_x=size_x
        self.size_y=size_y















from tkinter import *
import random
import math
from .SphereFrictionLeesEdwards import SphereFrictionLeesEdwards
from particleShearBase import CanvasPointsFrictionElasticityLeesEdwards
from .Ensemble import r_estimate
from .Ensemble import adjust_m
from particleShearBase import Force_register

class EnsembleFrictionLeesEdwards(CanvasPointsFrictionElasticityLeesEdwards):
    """Ensemble of particles of type  `particleShear.SphereFrictionLeesEdwards`, with Lees-Edwards boundary conditions

                This class creates and maintains a list of objects of type
                `particleShear.SphereFrictionLeesEdwards`. These objects are generically
                referred to as spheres and are stored in `particleShear.CanvasPoints.sphereList`.\n
                This class uses Lees-Edwards boundary conditions\n
                Class defined in subpackage particleShearObjects"""

    def __init__(self, size_x, size_y, N, packing_fraction=0.8, theCanvas=FALSE, doDrawing=FALSE, k=1,
                 nu=0.01,m=1,
                 bimodal_factor=1.4,
                 k_t=1,nu_t=0.01,mu=0.1):



        """Initialize self

                        - **parameters**\n
                        `size_x` Width of area to be used in pixels = micrometers for the simulation\n
                        `size_y` Height of area to be used in pixels = micrometers for the simulation\n
                        `N` The number of spheres to place
                        `packing_fraction` The packing fraction indicating the density; this is defined as the area occupied by the
                        non-compressed spheres as compared to the actual available area (size_x times size_y)
                        `theCanvas` possibility to transmit a `tkinter` `Canvas` object for graphical output\n
                        `doDrawing` Flag to indicate whether graphical output should be produced or not\n
                        `k` Central spring constant in (mg/s^2)/m of depth;
                        so F=-k*x*L with L the depth of the stack (the simulation is 2D) and x the compression\n
                        `nu` Central viscosity force constant in (mg/s)/m of depth. This is to have F=nu*v*L, L again the depth\n
                        `m` is the average mass per unit of depth in mg/m of depth\n
                        `bimodal_factor` is ratio of radii of the smaller and larger spheres in the bimodal distribution\n
                        `k_t` Tangential spring constant in (mg/s^2)/m of depth;
                        relevant for locked or permanent interfaces but not for slipping interfaces\n
                        `nu_t` Central viscosity force constant in (mg/s)/m of depth.
                         Relevant for frictionally locked and permanent interfaces but not slipping ones\n
                        `mu` Friction coefficient, describes the maximum interface force for non-permanent interfaces by
                         F_tangential_max = F_central*mu
                        """

        super(EnsembleFrictionLeesEdwards, self).__init__( size_x, size_y, theCanvas, False,
                                                          k, nu, m,
                                                          k_t=k_t,nu_t=nu_t,mu=mu)

        self.bimodal_upper = math.sqrt(bimodal_factor)
        """Upper multiplication factor for the bimodal sphere size distribution"""
        self.bimodal_lower = math.sqrt(1 / bimodal_factor)
        """Lower multiplication factor for the bimodal sphere size distribution"""

        self.N = N
        """Number of spheres"""
        self.packing_fraction = packing_fraction
        """Packing fraction: relative area occupied by the uncompressed spheres as compared to the available area"""

        self.doDrawing = doDrawing
        """Boolean: Should we draw the ensemble during simulation"""

        self.sphereList = []
        """List of the spheres"""
        r=self.size_y/2 # Default value if no spheres are specified

        if N > 0:
            r = r_estimate(self.size_x, self.size_y, packing_fraction, N,
                           bimodal_upper=self.bimodal_upper,
                           bimodal_lower=self.bimodal_lower)

        self.r_max = 0
        """Largest sphere radius"""
        for i in range(N):
            r_sphere = self.random_radius_bimodal(r)
            if r_sphere > self.r_max:
                self.r_max = r_sphere
            self.sphereList.append(SphereFrictionLeesEdwards("grey",
                                                     random.randrange(0, size_x),
                                                     random.randrange(0, size_y), r_sphere * 2,
                                                     adjust_m(m, r_sphere, r),
                                                     i, theCanvas, False,self, self.size_x, self.size_y))
        self.setShearInSpheres()
        self.setShearRateInSpheres()

        self.initiateGraphics()

    def particle_info(self):
        """Provide short description of particle type of inclusion into output file

                Method defined in `particleShear.EnsembleFrictionLeesEdwards` """
        return "Frictional spheres under Lees-Edwards boundary conditions\n" \
               "\tbimodal size distribution, equal probability, \n\tr1="+\
               str(self.r_max*0.8/1.2)+"r2="+str(self.r_max)+"\n\tidentical mass="+str(self.sphereList[0].m)+"mg"

    def random_radius_bimodal(self,r):
        """Choose randomly one of the two bimodal radius values.

        Method defined in `particleShear.EnsembleFrictionLeesEdwards`"""
        return r*(self.bimodal_lower+(self.bimodal_upper-self.bimodal_lower)*random.randint(0, 1))




from tkinter import *
import random
import math
from .Point import Point


# Basic moving point
class PointLeesEdwards(Point):
    """Define a moving point, under Lees-Edwards boundary conditions.

    Lees-Edwards boundary conditions are period boundary conditions where particles exiting on one face of the
     rectangular simulation area reappear on the opposing face; in addition, during such jumps, the x-position is adjusted to reflect
     applied shear (e.g. delta_x = delta_y * shear) and x-speed is adjusted to reflect applied shear rate (e.g.
     delta_v_x = delta_y * shear_rate). This is well illustrated on a number of websites or publications
     (for instance, Bertevas, Erwan & Fan, Xijun & I. Tanner, Roger. (2009).
     Simulation of the rheological properties of suspensions of oblate
     spheroidal particles in a Newtonian fluid. Rheologica Acta. 49. 53-73. 10.1007/s00397-009-0390-8. )\n\n
     Oriented-object implementation of the Lees-Edwards boundary conditions requires that the object is aware of
     the pertinent global parameters of the simulation in addition to its own local properties:
     size of the simulation rectangle, current shear rate, current shear."""
    def __init__(self, x, y,size_x=500,size_y=500,shear=0,shear_rate=0,use_lees_edwards=True):
        """Initialize self

        - **parameters**\n
        `x` The x-position of the point (assumed micrometers for the simulation)\n
        `y` The y-position of the point (assumed micrometers for the simulation)\n
        `size_x` Width of area to be used in pixels = micrometers for the simulation\n
        `size_y` Height of area to be used in pixels = micrometers for the simulation\n
        `shear` shear (affecting the x-coordinate; expressed relative, so dimension-less)\n
        `shear_rate` shear rate (affecting the x-movement; expressed relative, so units 1/s)"""
        super(PointLeesEdwards,self).__init__(x,y)

        self.use_lees_edwards=use_lees_edwards
        """Flag to indicate whether or not we use Lees-Edwards boundary conditions"""
        self.size_x=size_x
        """Width of area to be used in pixels = micrometers for the simulation"""
        self.size_y=size_y
        """Height of area to be used in pixels = micrometers for the simulation"""


        self.shear=shear
        """Current applied shear (relative, so dimension-less)"""
        self.shear_rate=shear_rate
        """Current shear rate (relative, so 1/s)"""









    def boundary_conditions(self, boundary_x, boundary_y):
        """Ensure location of the point within boundary conditions

               x should be >= 0 and < boundary_x\n
               y should be >= 0 and < boundary_y\n
               x will be adusted to satisfy this via addition or subtraction of multiples of boundary_x\n
               y will be adusted to satisfy this via addition or subtraction of multiples of boundary_y\n
               In addition to the parent method defined in `particleShear.Point`, here the corrections associated with
               the Lees-Edwards boundary conditions are taken into account by using
               `particleShear.PointLeesEdwards.lee_edwards_boundary_conditions` if
               `particleShear.PointLeesEdwards.use_lees_edwards` is set to True\n
                This method is defined in class `particleShear.PointLeesEdwards`"""
        self.size_x=boundary_x
        self.size_y=boundary_y



        if not self.use_lees_edwards:
            super(PointLeesEdwards,self).boundary_conditions(boundary_x,boundary_y)
        else:
            self.lee_edwards_boundary_conditions()




    def lee_edwards_closest_to_zero(self,x,y):
        """Return the closest match to zero (for x and y separately) under periodic transformations respecting the
        Lees-Edwards boudary conditions

        This method is defined in class `particleShear.PointLeesEdwards`
        """

        while y >= (self.size_y/2):
            y = y - self.size_y
            x = x - self.size_y * self.shear
        while y < (-self.size_y/2):
            y = y + self.size_y
            x = x + self.size_y * self.shear
        while x >= (self.size_x/2):
            x = x - self.size_x
        while x < (-self.size_x/2):
           x = x + self.size_x

        return [x, y]

    def lee_edwards_positive(self,x,y):
        """Return the smallest positive solution (for x and y separately) under periodic transformations respecting the
        Lees-Edwards boudary conditions

        This method is defined in class `particleShear.PointLeesEdwards`
        """
        while y >= self.size_y:
            y = y - self.size_y
            x = x - self.size_y * self.shear
        while y < 0:
            y = y + self.size_y
            x = x + self.size_y * self.shear
        while x >= self.size_x:
            x = x - self.size_x
        while x < 0:
           x = x + self.size_x

        return [x, y]


    def lee_edwards_relative_speed(self,theSphere):
        """Return relative speed using neighbor convention under Lees-Edwards boundary conditions

        For this, the relative speed for the nearest representation under repetition and shear is calculated. Adjustement
        in the x-speed is made only for transformation by multiples of the simulation area height\n
        This method is defined in class `particleShear.PointLeesEdwards`
        """

        vx_rel = theSphere.xspeed - self.xspeed
        vy_rel = theSphere.yspeed - self.yspeed

        y = theSphere.y - self.y
        while y >= (self.size_y / 2):
            y = y - self.size_y
            vx_rel = vx_rel - self.size_y * self.shear_rate
        while y < (-self.size_y / 2):
            y = y + self.size_y
            vx_rel = vx_rel + self.size_y * self.shear_rate


        return [vx_rel, vy_rel]

    def relative_speed(self,theSphere):
        """Relative speed of the two spheres evaluated by the difference of speed of their centers.

        If `particleShear.PointLeesEdwards.use_lees_edwards` is set to True, this
        method uses `particleShear.PointLeesEdwards.lee_edwards_relative_speed`; otherwise,
        it uses `particleShear.Point.relative_speed` defined in the parent class `particleShear.Point`\n
        This method is defined in class `particleShear.PointLeesEdwards`
        """
        if self.use_lees_edwards:
            return self.lee_edwards_relative_speed(theSphere)
        else:
            return super(PointLeesEdwards,self).relative_speed(theSphere)






    def lee_edwards_boundary_conditions(self):
        """Ensure location of the point within boundary conditions

        x should be >= 0 and < boundary_x\n
        y should be >= 0 and < boundary_y\n
        x will be adusted to satisfy this via addition or subtraction of multiples of boundary_x\n
        y will be adusted to satisfy this via addition or subtraction of multiples of boundary_y\n
        In addition, adjusting y is coupled with a change in x and speed in x-direction
        to reflect shear, respectively shear rate, according to the Lees-Edwards boundary conditions.\n
        This method is defined in class `particleShear.PointLeesEdwards`."""


        old_x = self.x
        old_y = self.y

        new_pos = self.lee_edwards_positive(self.x,self.y)

        # Adjust speed

        self.xspeed = self.xspeed + (new_pos[1]-self.y)*self.shear_rate

        self.x = new_pos[0]
        self.y = new_pos[1]

        if self.doDrawing:
            self.theCanvas.move(self.shape, self.x - old_x, self.y - old_y)

    def lee_edwards_relative_vector(self,theSphere):
        """Return a relative vector to the Sphere, using smallest x- and y- distances resulting under Lees-Edwards
        periodic boundary condition transforms.

        This is achieved by applying
        `particleShear.PointLeesEdwards.lee_edwards_closest_to_zero` to the actual delta x and delta y vector.\n
        This method is defined in class `particleShear.PointLeesEdwards`.
         """
        pos = theSphere.coordinates()
        x = pos[0]
        y = pos[1]
        delta_x = x - self.x
        delta_y = y - self.y
        return self.lee_edwards_closest_to_zero(delta_x, delta_y)


    def d(self,theSphere):
        """Return the distance to another sphere under Lees-Edwards boundary conditions.

        This is the shortest distance
        among periodic repetitions, as evaluated as the absolute length of the delta vector returned by
        `particleShear.PointLeesEdwards.lee_edwards_relative_vector`.
        If `particleShear.PointLeesEdwards.use_lees_edwards` is False, the parent method
        `particleShear.Point.d` of class `particleShear.Point` is used instead.\n
        This method is defined in class `particleShear.PointLeesEdwards`."""
        if not self.use_lees_edwards:
            return super(PointLeesEdwards,self).d(theSphere)

        newDelta = self.lee_edwards_relative_vector(theSphere)
        return math.sqrt(newDelta[0] * newDelta[0] + newDelta[1] * newDelta[1])
    # To get a normal unit vector to another sphere

    def n(self, theSphere):
        """Return normal unit vector pointing towards another sphere, using Lees-Edward boundary conditions for
        determining the shortest connection among the periodic representations.


        Uses `particleShear.PointLeesEdwards.lee_edwards_relative_vector` to determine the shortest distance among the
        peridic representations under the Lees-Edwards boundary conditions.
        If `particleShear.PointLeesEdwards.use_lees_edwards` is False, the parent method
        `particleShear.Point.n` of class `particleShear.Point` is used instead.\n
        This method is defined in class `particleShear.PointLeesEdwards`.
        """


        if not self.use_lees_edwards:
            return super(PointLeesEdwards,self).n(theSphere)

        d = self.d(theSphere)
        if d > 0:
            rel=self.lee_edwards_relative_vector(theSphere)
            return ([rel[0] / d, rel[1] / d])
        if d == 0:
            angle = random.random()* 2 * math.pi
            return [math.cos(angle), math.sin(angle)]


    def transmit_lees_edwards_parameters(self,theSphere):
        """Transmit the parameters pertaining to the Lees-Edwards boundary conditions to another sphere (that is,
        object deriving from `particleShear.PointLeesEdwards`).

        If `particleShear.PointLeesEdwards.use_lees_edwards` is set to True, transmit the instance variables regarding
        canvas size (`particleShear.PointLeesEdwards.size_x`,`particleShear.PointLeesEdwards.size_y`), and shear regime
        (`particleShear.PointLeesEdwards.shear`,`particleShear.PointLeesEdwards.shear_rate`,
        `particleShear.PointLeesEdwards.use_lees_edwards`) to the given target sphere.
        Else, only set the `particleShear.PointLeesEdwards.use_lees_edwards` field in the
        target sphere to False\n
        This method is defined in class `particleShear.PointLeesEdwards`"""
        if self.use_lees_edwards:
            theSphere.size_x = self.size_x
            theSphere.size_y = self.size_y
            theSphere.shear_rate = self.shear_rate
            theSphere.shear = self.shear
            theSphere.use_lees_edwards = self.use_lees_edwards
        else:
            if hasattr(theSphere,"use_lees_edwards"):
                theSphere.use_lees_edwards=False
        return theSphere


    def cool(self, f=0.8):
        """Cool by decreasing local speed

        Cool relative to fine-grained local movement (x-speed of the fluid is expected to be
        (y-size_y/2)*shear_rate), decrease the delta to this expected speed rather than
        diminishing the absolute speed. This is used to stabilize the ensemble at the given
        shear; physically, this reflects friction with the pore or interstitial fluid.\n
        This method is defined in class `particleShear.PointLeesEdwards`"""
        x_speed_local = (self.y-self.size_y/2)*self.shear_rate

        #self.xspeed = x_speed_local+(self.xspeed-x_speed_local) * f
        self.xspeed = x_speed_local + (self.xspeed - x_speed_local) * f
        self.yspeed = self.yspeed * f






